from __future__ import annotations

import logging
import uuid
from dataclasses import fields
from enum import Enum
from typing import Any, Dict, List, NamedTuple, Optional, Tuple

from blinker import Signal

import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.adapters import AdapterListProtocol, BaseAdapter, ready_signal
from ieee_2030_5.adapters.timeadapter import TimeAdapter
from ieee_2030_5.config import InvalidConfigFile
from ieee_2030_5.data.indexer import add_href, get_href_filtered
from ieee_2030_5.models.sep import DERProgram
from ieee_2030_5.types_ import StrPath

_log = logging.getLogger(__name__)

__all__: List[str] = [
    "DERControlAdapter",
    "DERProgramAdapter",
    "DERCurveAdapter",
    "DERAdapter"
]

class CreateStatus(Enum):
    Error = "Error"
    Created = "Created"
    Updated = "Updated"

class CreateResponse(NamedTuple):
    data: Any
    href: str
    index: int = -1
    status: str = CreateStatus.Created.value
    
    @property
    def statusint(self) -> int:
        if self.status == CreateStatus.Created.value:
            return 201
        elif self.status == CreateStatus.Updated.value:
            return 204
        elif self.status == CreateStatus.Error:
            return 400
        
    


class _DERCurveAdapter(BaseAdapter, AdapterListProtocol):
    
    def __init__(self) -> None:
        super().__init__()
        self._curves: List[m.DERCurve] = []
    

    def __initialize__(self, sender):
        """Initialize the DERCurve objects based upon the BaseAdapter.__device_configuration__"""

        if BaseAdapter.device_configs() is None:
            raise ValueError("Initialize BaseAdapter before initializing the Curves.")

        curves_cfg = BaseAdapter.server_config().curves

        for index, curve_cfg in enumerate(curves_cfg):

            der_curve = m.DERCurve(**curve_cfg.__dict__)
            der_curve.href = hrefs.curve_href(index)
            self._curves.append(der_curve)
            
        ready_signal.send(self)

    def fetch_edev_all(self) -> List[m.DERCurve]:
        """Fetch all of the 2030.5 DERCurve objects as a list."""
        return self._curves

    def fetch_list(self, start=0, after=0, limit=0) -> m.DERCurveList:
        """Fetch a 2030.5 DERCurve object as a DERCurveList"""
        curve_list = m.DERCurveList(href=hrefs.curve_href())
        curve_list.DERCurve = self._curves
        curve_list.all = len(self._curves)
        curve_list.results = len(self._curves)
        return curve_list


DERCurveAdapter = _DERCurveAdapter()
ready_signal.connect(DERCurveAdapter.__initialize__, BaseAdapter)
#BaseAdapter.after_initialized.connect(DERCurveAdapter.__initialize__)


class _DERControlAdapter(BaseAdapter, AdapterListProtocol):

    def __init__(self) -> None:
        super().__init__()
        self._default_control: m.DefaultDERControl
        self._controls: List[m.DERControl] = []

    @staticmethod
    def fetch_default() -> m.DefaultDERControl:
        dderc = get_href(hrefs.get_dderc_href())

        if dderc is None:
            derbase = m.DERControlBase(opModConnect=True, opModEnergize=False)

            # Defaults from Jakaria on 1/26/2023
            dderc = m.DefaultDERControl(href=hrefs.get_dderc_href(),
                                        mRID=str(uuid.uuid4()),
                                        description="Default DER Control Mode",
                                        setESDelay=300,
                                        setESLowVolt=0.917,
                                        setESHighVolt=1.05,
                                        setESLowFreq=59.5,
                                        setESHighFreq=60.1,
                                        setESRampTms=300,
                                        setESRandomDelay=0,
                                        DERControlBase=derbase)
            add_href(dderc.href, dderc)

        return dderc

    @staticmethod
    def store_default(dderc: m.DefaultDERControl):
        add_href(hrefs.get_dderc_href(), dderc)

    def __initialize__(self, sender):
        """Initialize the DERControl and DERControlBase objects based upon the BaseAdapter.__server_configuration__"""
        config = DERControlAdapter.server_config().controls

        for index, ctl in enumerate(config):
            
            # Create a new DERControl and DERControlBase and initialize as much as possible
            base_control: m.DERControlBase = BaseAdapter.build_instance(m.DERControlBase, ctl.base)

            control: m.DERControl = BaseAdapter.build_instance(m.DERControl, ctl.__dict__)
            control.href = hrefs.get_derc_href(index=index)
            #control.mRID = f"MYCONTROL{index}"
            control.DERControlBase = base_control
            
            self._controls.append(control)
        
        ready_signal.send(self)
        
    def fetch_edev_all(self) -> List[m.DERControl]:
        return self._controls
    
    def fetch_list(self, start: int = 0, after: int = 0, limit: int = 0) -> m.DERControlList:
        ctrl_list = m.DERControlList(href=hrefs.get_derc_href(),
                                     DERControl=self._controls,
                                     all=len(self._controls),
                                     results=len(self._controls))
        return ctrl_list

    @staticmethod
    def initialize_from_storage():
        dercs_href = hrefs.get_derc_href(hrefs.NO_INDEX)
        der_controls = get_href_filtered(dercs_href)
        DERControlAdapter.__count__ = len(der_controls)

    @staticmethod
    def create_from_parameters(**kwargs) -> m.DERControl:
        """Create and store a new DERControl object from the passed kwargs
        
        If invalid parameters are passed then a ValueError should be raised.        
        """
        control = m.DERControl()
        base_control = m.DERControlBase()
        control.DERControlBase = base_control
        field_list = fields(control)
        for k, v in kwargs.items():
            for f in field_list:
                if k == f.name:
                    setattr(control, k, v)
            for f in fields(base_control):
                if k == f.name:
                    setattr(base_control, k, v)

    @staticmethod
    def build_der_control(**kwargs) -> m.DERControl:
        """ Build a DERControl object from the passed parameters.

        Args:
            **params: The parameters passed in from the web application.
        """
        control = m.DERControl()
        base_control = m.DERControlBase()
        control.DERControlBase = base_control
        field_list = fields(control)
        for k, v in kwargs.items():
            for f in field_list:
                if k == f.name:
                    setattr(control, k, v)
            for f in fields(base_control):
                if k == f.name:
                    setattr(base_control, k, v)

        return control

    @staticmethod
    def store_single(der_control: m.DERControl | m.DefaultDERControl):
        if not der_control.mRID:
            der_control.mRID = str(uuid.uuid4())

        if not der_control.href:
            der_control.href = hrefs.get_derc_href(DERControlAdapter.__count__)
            add_href(der_control.href, der_control)
            DERControlAdapter.__count__ += 1

        add_href(der_control.href, der_control)

    @staticmethod
    def load_from_storage() -> Tuple[List[m.DERControl], m.DefaultDERControl]:
        der_controls, default_der_control = get_href_filtered(hrefs.get_derc_href(hrefs.NO_INDEX))
        return der_controls, default_der_control

    @staticmethod
    def get_all() -> List[m.DERControl]:
        """ Retrieve a list of dataclasses for DERControl for the """
        return list(filter(lambda a: isinstance(a, m.DERControl), get_href_filtered("")))

    @staticmethod
    def load_from_yaml_file(yaml_file: StrPath) -> Tuple[List[m.DERControl], m.DefaultDERControl]:
        """Load from a configuration yaml file

        The yaml file must have a list of DERControls dictionary items and a default str that matches
        with one of the DERControlName.  This method loops over the DERControls and
        creates a DERControlBase with the parameters in the config file.

        The validations that could throw errors are:

            - Required fields are DERControlName, mRID, description
            - Required no-duplicates in mRID and DERControlName

        Returns List[DERControls], DefaultDERControl
        """
        if isinstance(yaml_file, str):
            yaml_file = Path(yaml_file)
        yaml_file = yaml_file.expanduser()
        data = yaml.safe_load(yaml_file.read_text())
        default = data.get('default', None)
        if not default:
            raise ValueError(
                f"A default DERControl must be specified in {self.DERControlListFile}")

        # DERControls in the yaml file should be an array of DERControl instances.
        if 'DERControls' not in data or \
                not isinstance(data.get('DERControls'), list):
            raise ValueError(
                f"DERControls must be a list within the yaml file {self.DERControlListFile}")

        default_derc: m.DefaultDERControl = None
        derc_list: List[m.DERControl] = []
        derc_names: Dict[str, str] = {}
        mrids: Dict[str, str] = {}
        for index, derc in enumerate(data["DERControls"]):
            required = 'mRID', 'DERControlName', 'description'
            for req in required:
                if req not in derc:
                    raise ValueError(f"{req}[{index}] does not have a '{req}' in {yaml_file}.")
            name = derc.pop('DERControlName')
            mrid = derc.pop('mRID')
            description = derc.pop('description')
            if mrid in mrids:
                raise ValueError(f"Duplicate mrid {mrid} found in {yaml_file}")
            if name in derc_names:
                raise ValueError(f"Duplicate name {name} found in {yaml_file}")

            base_field_names = [fld.name for fld in fields(m.DERControlBase)]
            base_dict = {ctl: derc[ctl] for ctl in derc if ctl in base_field_names}
            control_fields = {ctl: derc[ctl] for ctl in derc if not ctl in base_dict.keys()}
            control_base = m.DERControlBase(**base_dict)
            item = m.DERControl(mRID=mrid,
                                description=description,
                                DERControlBase=control_base,
                                **control_fields)
            item.href = hrefs.get_derc_href(index)

            # TODO: Figure out if we need a global default or if that
            # is specific to the DERProgram construct.
            # Build a default der control
            # if name == default:
            #     try:
            #         default_derc = m.DefaultDERControl(mRID=mrid,
            #                                            description=description,
            #                                            DERControlBase=control_base,
            #                                            **control_fields)
            #         default_derc.href = hrefs.get_derc_default_href()
            #         add_href(default_derc.href, default_derc)
            #     except TypeError as ex:
            #         raise InvalidConfigFile(f"{yaml_file}\ndefault config parameter {ex.args[0]}")

            add_href(item.href, item)
            derc_list.append(item)

        # if not default_derc:
        #     raise InvalidConfigFile(f"{yaml_file} no DefaultDERControl specified.")
        if not len(derc_list) > 0:
            raise InvalidConfigFile(f"{yaml_file} must have at least one DERControl specified.")

        return derc_list, None    # default_derc

DERControlAdapter = _DERControlAdapter()
ready_signal.connect(DERControlAdapter.__initialize__, DERCurveAdapter)


class _DERProgramAdapter(BaseAdapter, AdapterListProtocol):
    
    def __init__(self) -> None:
        super().__init__()
        
        self._der_programs: List[DERProgram] = []
        self._der_default_control: Dict[int, m.DefaultDERControl] = {}
        self._der_control: Dict[int, List[m.DERControl]] = {}
        TimeAdapter.tick.connect(self._time_updated)
        
    def _time_updated(self, timestamp):
        
        for indx, ctrl_list in self._der_control.items():
            event_removed = []
            for ctrl_index, ctrl in enumerate(ctrl_list):
                if not ctrl.EventStatus:
                    if ctrl.interval.start > timestamp and timestamp < ctrl.interval.start + ctrl.interval.duration:
                        ctrl.EventStatus = m.EventStatus(currentStatus=1, dateTime=timestamp, potentiallySuperseded=False, reason="Active") 
                    else:
                        ctrl.EventStatus = m.EventStatus(currentStatus=0, dateTime=timestamp, potentiallySuperseded=False, reason="Scheduled") 

                # Active control
                if ctrl.interval.start < timestamp and timestamp < ctrl.interval.start + ctrl.interval.duration:
                    # TODO send active control event
                    if ctrl.EventStatus.currentStatus == 0:
                        ctrl.EventStatus.currentStatus = 1 # Active
                        ctrl.EventStatus.dateTime = timestamp
                        ctrl.EventStatus.reason = f"Control event active {ctrl.mRID}"
                        
                elif timestamp > ctrl.interval.start + ctrl.interval.duration:
                    if ctrl.EventStatus.currentStatus == 1: # Only remove if we were active (otherwise we need to delete differently)
                        event_removed.append(ctrl_index)
                    
            for i in sorted(event_removed, reverse=True):
                ctrl_list.pop(i)
                    
                    
                    
    
    def __initialize__(self, sender):
        """Initialize the DERProgram objects based upon the BaseAdapter.__server_configuration__"""
        cfg_programs = BaseAdapter.__server_configuration__.programs
        cfg_der_controls = BaseAdapter.__server_configuration__.controls
        _log.debug("Update m.DERPrograms' adding links to the different program pieces.")

        der_controls = DERControlAdapter.fetch_edev_all()
        der_curves = DERCurveAdapter.fetch_edev_all()
        # Initialize "global" m.DERPrograms href lists, including all the different links to
        # locations for active, default, curve and control lists.
        for index, program_cfg in enumerate(cfg_programs):
            # The configuration contains a mapping to control lists so when
            # building the DERProgram object we need to remove it from the paramters before
            # initialization.
            params = program_cfg.__dict__.copy()
            del params['default_control']
            del params['controls']
            del params['curves']
            
            program = m.DERProgram(**params)
            program.description = program_cfg.description
            program.primacy = program_cfg.primacy
            # program.version = program_cfg.version

            # TODO Fix this!
            # program.mRID =
            # mrid = program_cfg.get('mrid')
            # if mrid is None or len(mrid.trim()) == 0:
            #     program.mRID = f"program_mrid_{index}"
            # program_cfg['mrid'] = program.mRID

            program.href = hrefs.get_program_href(index)
            # program.ActiveDERControlListLink = m.ActiveDERControlListLink(hrefs..der_href())

            try:
                der_ctl = next(
                    filter(lambda d: d.description == program_cfg.default_control, der_controls))
                der_cfg = next(
                    filter(lambda d: d.description == program_cfg.default_control,
                           cfg_der_controls))
            except StopIteration:
                raise InvalidConfigFile(
                    f"Section program: {program_cfg.description} default control {program_cfg.default_control} not found!"
                )
            else:
                default_ctl: m.DefaultDERControl = BaseAdapter.build_instance(
                    m.DefaultDERControl, der_cfg.__dict__)
                default_ctl.href = hrefs.get_derc_default_href(index)
                #default_ctl.mRID = der_ctl.mRID + " default"
                default_ctl.setESDelay = 20
                default_ctl.DERControlBase = der_ctl.DERControlBase

                program.DefaultDERControlLink = m.DefaultDERControlLink(href=default_ctl.href)
                self._der_programs.append(program)

            # der_control_list = m.DERControlList(href=hrefs.get_program_href(index, hrefs.DERC))

            # for ctl_description in program_cfg.controls:
            #     try:
            #         derc = next(filter(lambda d: d.description == ctl_description, der_controls))
            #     except StopIteration:
            #         raise InvalidConfigFile(
            #             f"Section program: {program_cfg.description} control {ctl_description} not found!"
            #         )
            #     else:
            #         der_control_list.DERControl.append(derc)

            # add_href(der_control_list.href, der_control_list)

            # der_curve_list = m.DERCurveList(href=hrefs.get_program_href(index, hrefs.CURVE))

            # for curve_description in program_cfg.curves:
            #     try:
            #         der_curve = next(
            #             filter(lambda d: d.description == curve_description, der_curves))
            #     except StopIteration:
            #         raise InvalidConfigFile(
            #             f"Section program: {program_cfg.description} curve {curve_description} not found!"
            #         )
            #     else:
            #         der_curve_list.DERCurve.append(der_curve)

            # der_curve_list.all = len(der_curve_list.DER)
            ready_signal.send(self)

    def fetch_edev_all(self) -> List:
        return self._der_programs
    
    def fetch_list(self, start: int = 0, after: int = 0, limit: int = 0) -> m.DERProgramList:
        program_list = m.DERProgramList(href=hrefs.der_program_href(), DERProgram=self._der_programs,
                                        all=len(self._der_programs),
                                        results=len(self._der_programs))
        return program_list
    
    def fetch_der_default_control(self, program_index: int) -> m.DefaultDERControl:
        return self._der_default_control.get(program_index, m.DefaultDERControl())
    
    def fetch_der_control_list(self, program_index: int, start: int = 0, after: int = 0, limit: int = 0) -> m.DERControlList:
        der_controls = self.get_der_controls(program_index)
        all=len(der_controls)
        return m.DERControlList(href=hrefs.der_program_href(program_index, hrefs.DERProgramSubType.DERControlListLink), all=all, results=all,
                                DERControl=der_controls)
        
    def fetch_der_active_control_list(self, program_index: int, int, start: int = 0, after: int = 0, limit: int = 0) -> m.DERControlList:
        der_control_list = m.DERControlList(href=hrefs.der_program_href(program_index, hrefs.DERProgramSubType.ActiveDERControlListLink),
                                            DERControl=self.get_der_active_controls(program_index=program_index))
        
        der_control_list.all = len(der_control_list.DERControl)
        der_control_list.results = len(der_control_list.DERControl)
        return der_control_list
    
    def get_der_active_controls(self, program_index: int) -> List[m.DERControl]:
        lst: List[m.DERControl] = []
        for ctrl in self._der_control.get(program_index, []):
            if ctrl.EventStatus.currentStatus == 1: # Active
                lst.append(ctrl)
        return lst
    
    def get_default_der_control(self, program_index) -> m.DefaultDERControl:
        return self._der_default_control.get(program_index)
    
    def get_der_controls(self, program_index) -> List[m.DERControl]:
        return self._der_control.get(program_index, [])
    
    def create_default_der_control(self, program_index, data: m.DefaultDERControl) -> CreateResponse:
        if program_index in self._der_default_control:
            status = CreateStatus.Updated.value
        else:
            status = CreateStatus.Created.value
        data.href = hrefs.der_program_href(program_index, hrefs.DERProgramSubType.DefaultDERControlLink)
        self._der_default_control[program_index] = data
        return CreateResponse(data, href=data.href, status=status)
    
    def create_der_control(self, program_index, data: m.DERControl)  -> CreateResponse:
        found_derc_index = -1
        
        if program_index in self._der_control:
            for index, derc in enumerate(self._der_control[program_index]):
                if derc.mRID == data.mRID:
                    found_derc_index = index
                    break
        else:
            self._der_control[program_index] = []
        
        data.replyTo = hrefs.der_program_href(program_index, hrefs.DERProgramSubType.DERControlReplyTo) 
        if found_derc_index == -1:
            status = CreateStatus.Created.value
            found_derc_index = len(self._der_control[program_index])
            self._der_control[program_index].append(data)
        else:
            status = CreateStatus.Updated.value
            self._der_control[program_index][found_derc_index] = data
    
        data.href = hrefs.der_program_href(program_index, hrefs.DERProgramSubType.DERControlListLink, found_derc_index)
        
        return CreateResponse(data, href=data.href, status=status)
    
    def fetch_by_mrid(self, mRID: str) -> Optional[m.DERProgram]:
        try:
            der_program = next(filter(lambda x: x.mRID == mRID, self._der_programs))
        except StopIteration:
            der_program = None
            
        return der_program
    
    def fetch_at(self, index: int) -> m.DERProgram:
        return self._der_programs[index]
        
    def create(self, data: m.DERProgram) -> CreateResponse:
        
        found_index = -1
        if data.mRID:
            for index, ele in enumerate(self._der_programs):
                if ele.mRID == data.mRID:
                    found_index = index
                    break
                
        def update_hrefs(index: int, data: m.DERProgram):
            data.href = hrefs.der_program_href(index)
            
            if active := self.get_der_active_controls(index):
                data.ActiveDERControlListLink = m.ActiveDERControlListLink(hrefs.der_program_href(index, hrefs.DERProgramSubType.ActiveDERControlListLink), 
                                                                           all=len(active))
            if ctrl := self.get_der_controls(index):
                data.DERControlListLink = m.DERControlListLink(hrefs.der_program_href(index, hrefs.DERProgramSubType.DERControlListLink),
                                                               all=len(ctrl))
            
            if dctrl := self.get_default_der_control(index):
                data.DefaultDERControlLink = m.DefaultDERControlLink(hrefs.der_program_href(index, hrefs.DERProgramSubType.DefaultDERControlLink))
            
            
        
        if found_index != -1:
            self._der_programs[found_index] = data
            the_index = found_index            
            response = CreateResponse(data, index=found_index, href=data.href, status=CreateStatus.Updated.value)
        else:
            new_index = len(self._der_programs)
            self._der_programs.append(data)
            the_index = new_index    
            response = CreateResponse(data, index=new_index, href=data.href)

        update_hrefs(the_index, data)
        
        return response

    @staticmethod
    def build(**kwargs) -> m.DERProgram:
        """ Build a DERProgram from the passed kwargs

        kwarg variables:
            der_control_checked<int> - will be passed with the value of the href for the checked control.
            default_der_control - Is the default der control that should be used when no controls are active.
        """
        program = m.DERProgram()

        kwargs = populate_from_kwargs(program, **kwargs)

        href_default = kwargs.pop('default_der_control', None)
        hrefs_controls = [kwargs[k] for k in kwargs if k.startswith('der_control_checked')]

        program.DefaultDERControlLink = m.DefaultDERControlLink(href=href_default)
        # m.DERControlList
        program.DERControlListLink = hrefs.get_program_href()

        return program

DERProgramAdapter = _DERProgramAdapter()
ready_signal.connect(DERProgramAdapter.__initialize__, DERControlAdapter)


class _DERAdapter(BaseAdapter, AdapterListProtocol):
    def __init__(self) -> None:
        super().__init__()
        
        self._edev_ders: Dict[int, List[m.DER]] = {}
        self._der_capabilities: Dict[int, List[m.DERCapability]] = {}
        self._der_settings: Dict[int, List[m.DERSettings]] = {}
        self._der_status: Dict[int, List[m.DERStatus]] = {}
        self._der_availabilites: Dict[int, List[m.DERAvailability]] = {}
        self._der_current_program: Dict[int, List[m.DERProgram]] = {}
        self._edev_der_settings: Dict[int, List[m.DERSettings]] = {}
        
    def __initialize__(self, sender):
        # TODO: Load ders
        cfg = BaseAdapter.server_config()
        
    def create(self, edev_index: int, der_index: int,  modesSupported: str, deviceType: int) -> m.DER:
        
        der = m.DER(href=hrefs.edev_der_href(edev_index=edev_index, der_index=der_index))
        der.DERCapabilityLink = m.DERCapabilityLink(hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.Capability))
        der.DERSettingsLink = m.DERSettingsLink(hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.Settings))
        der.DERStatusLink = m.DERStatusLink(hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.Status))
        der.DERAvailabilityLink = m.DERAvailabilityLink(hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.Availability))
        der.CurrentDERProgramLink = m.CurrentDERProgramLink(hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.CurrentProgram))
        if not self._edev_ders.get(edev_index):
            self._edev_ders[edev_index] = []
            
        self._edev_ders[edev_index].append(der)
        
        if not self._der_capabilities.get(edev_index):
            self._der_capabilities[edev_index] = []
        self._der_capabilities[edev_index].append(m.DERCapability(hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.Capability),
                                                             modesSupported=modesSupported,
                                                             type=deviceType))
        
        if not self._der_settings.get(edev_index):
            self._der_settings[edev_index] = []
            
        self._der_settings[edev_index].append(m.DERSettings(href=hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.Settings)))
        
        if not self._der_status.get(edev_index):
            self._der_status[edev_index] = []
        
        self._der_status[edev_index].append(m.DERStatus(href=hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.Status)))
        
        if not self._der_availabilites.get(edev_index):
            self._der_availabilites[edev_index] = []
        
        self._der_availabilites[edev_index].append(m.DERAvailability(href=hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.Availability)))
        
        if not self._der_current_program.get(edev_index):
            self._der_current_program[edev_index] = []
        
        self._der_current_program[edev_index].append(m.DERProgram(href=hrefs.der_sub_href(edev_index=edev_index, index=der_index, subtype=hrefs.DERSubType.CurrentProgram)))
        
        return der
    
    def fetch_edev_all(self, edev_index: int) -> List:
        return self._edev_ders.get(edev_index, [])
    
    def fetch_list(self, edev_index: int, start: int = 0, after: int = 0, limit: int = 0) -> m.DERList:
        
        try:
            der_list = m.DERList(href=hrefs.der_sub_href(edev_index), 
                                all=len(self._edev_ders[edev_index]), results=len(self._edev_ders[edev_index]), DER=self._edev_ders[edev_index])
        except KeyError:
            der_list = m.DERList(href=hrefs.der_sub_href(edev_index), 
                                 all=0, results=0, DER=[])
        return der_list
    
    def fetch_at(self, edev_index: int, der_index: int) -> m.DER:
        return self._edev_ders[edev_index][der_index]
    
    def fetch_settings_at(self, edev_index: int, der_index: int) -> m.DERSettings:
        return self._edev_der_settings[edev_index][der_index]
    
    def store_settings_at(self, edev_index: int, der_index: int, settings: m.DERSettings):
        self._edev_der_settings[edev_index][der_index] = settings
    
    def fetch_status_at(self, edev_index: int, der_index: int) -> m.DERStatus:
        return self._der_status[edev_index][der_index]
    
    def store_status_at(self, edev_index: int, der_index: int, status: m.DERStatus):
        self._edev_der_status[edev_index][der_index] = status
    
    def fetch_capability_at(self, edev_index: int, der_index: int) -> m.DERCapability:
        return self._der_capabilities[edev_index][der_index]
    
    def store_capability_at(self, edev_index: int, der_index: int, capability: m.DERCapability):
        self._der_capabilities[edev_index][der_index] = capability
    
    def fetch_availibility_at(self, edev_index: int, der_index: int) -> m.DERAvailability:
        return self._der_availabilites[edev_index][der_index]
    
    def store_availibility_at(self, edev_index: int, der_index: int, availability: m.DERAvailability):
        self._der_availabilites[edev_index][der_index] = availability
    
    def fetch_current_program_at(self, edev_index: int, der_index: int) -> m.DERProgram:
        return self._der_current_program[edev_index][der_index]
    
    def store_current_program_at(self, edev_index: int, der_index: int, current_program: m.DERProgram):
        self._der_current_program[edev_index][der_index] = current_program
    
    def store(self, parsed_edev: hrefs.EdevHref, data) -> int:
        if parsed_edev.der_sub == hrefs.DERSubType.Availability.value:
            self._der_availabilites[parsed_edev.edev_index][parsed_edev.der_index] = data
        elif parsed_edev.der_sub == hrefs.DERSubType.Capability.value:
            self._der_capabilities[parsed_edev.edev_index][parsed_edev.der_index] = data
        elif parsed_edev.der_sub == hrefs.DERSubType.Status.value:
            self._der_status[parsed_edev.edev_index][parsed_edev.der_index] = data
        elif parsed_edev.der_sub == hrefs.DERSubType.Settings.value:
            self._der_settings[parsed_edev.edev_index][parsed_edev.der_index] = data
        else:
            raise ValueError(data)
        
        return 200
    
    def get_list(self, edev_index: int):
        return self._edev_ders[edev_index]
    
    
DERAdapter = _DERAdapter()
ready_signal.connect(DERAdapter.__initialize__, BaseAdapter)



if __name__ == '__main__':
    from pathlib import Path

    import yaml

    from ieee_2030_5.__main__ import get_tls_repository
    from ieee_2030_5.config import ServerConfiguration
    
    cfg_pth = Path("/home/os2004/repos/gridappsd-2030_5/config.yml")
    cfg_dict = yaml.safe_load(cfg_pth.read_text())

    config = ServerConfiguration(**cfg_dict)

    tls_repo = get_tls_repository(config, False)

    BaseAdapter.initialize(config, tls_repo)    

    print(DERCurveAdapter.fetch_edev_all())
    print(DERCurveAdapter.fetch_list())


import json
from typing import Optional

from flask import Flask, Response, render_template, request

import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.adapters.der import DERAdapter, DERProgramAdapter
from ieee_2030_5.adapters.enddevices import EndDeviceAdapter
from ieee_2030_5.adapters.fsa import FSAAdapter
from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.config import ServerConfiguration
from ieee_2030_5.server.server_constructs import EndDevices
from ieee_2030_5.utils import dataclass_to_xml, xml_to_dataclass


class AdminEndpoints:
    def __init__(self, app: Flask, tls_repo: TLSRepository, config: ServerConfiguration):
        self.tls_repo = tls_repo
        self.server_config = config
        
        app.add_url_rule("/admin", view_func=self._admin)
        app.add_url_rule("/admin/enddevices/<int:index>", view_func=self._admin_enddevices)    
        app.add_url_rule("/admin/enddevices", view_func=self._admin_enddevices)
        app.add_url_rule("/admin/end-device-list", view_func=self._admin_enddevice_list)
        app.add_url_rule("/admin/program-lists", view_func=self._admin_der_program_lists)
        app.add_url_rule("/admin/lfdi", endpoint="admin/lfdi", view_func=self._lfdi_lists)
        app.add_url_rule("/admin/edev/<int:edev_index>/ders/<int:der_index>/current_derp", view_func=self._admin_der_update_current_derp, methods=['PUT', 'GET'])
        app.add_url_rule("/admin/ders/<int:edev_index>", view_func=self._admin_ders)
        
        app.add_url_rule("/admin/edev/<int:edevid>/fsa", view_func=self._admin_edev_fsa)
        app.add_url_rule("/admin/derp/<int:index>/derc/<int:control_index>",  methods=['GET', 'PUT'], view_func=self._admin_derp_derc)
        app.add_url_rule("/admin/derp/<int:index>/derc",  methods=['GET', 'POST'], view_func=self._admin_derp_derc)
        app.add_url_rule("/admin/derp/<int:index>/dderc",  methods=['GET', 'PUT'], view_func=self._admin_derp_derc)
        app.add_url_rule("/admin/derp",  methods=['GET', 'POST'], view_func=self._admin_derp)
        #app.add_url_rule("/admin/derp/<int:index>",  methods=['GET', 'POST'], view_func=self._derp)
        #app.add_url_rule("/admin/derp/<int:index>/derc", methods=['GET', 'POST'], view_func=self._derp_derc)
        
    def _admin_der_program_lists(self) -> Response:
        return Response(dataclass_to_xml(DERProgramAdapter.fetch_list()))
    
    def _admin_ders(self, edev_index: int) -> Response:
        return Response(dataclass_to_xml(DERAdapter.fetch_list(edev_index=edev_index)))
    
    def _der_settings(self, edev_index: int, der_index: int):
        return Response(dataclass_to_xml(DERAdapter.fetch_settings_at(edev_index=edev_index, der_index=der_index)))
        
    def _admin_der_update_current_derp(self, edev_index: int, der_index: int):
        if request.method == 'PUT':
            data: m.DERProgram = xml_to_dataclass(request.data.decode('utf-8'))
            if not isinstance(data, m.DERProgram):
                return Response(status=400)
            
            if data.mRID:
                program = DERProgramAdapter.fetch_by_mrid(data.mRID)
                response_status = 200
                if not program:
                    program = DERProgramAdapter.create(data).data
                    response_status = 201
            else:
                program = DERProgramAdapter.create(data).data
                response_status = 201
            
            der = DERAdapter.fetch_at(edev_index, der_index)
            der.CurrentDERProgramLink = m.CurrentDERProgramLink(program.href)
            return Response(dataclass_to_xml(program), status=response_status)
        else:
            der = DERAdapter.fetch_at(edev_index, der_index)
            if der.CurrentDERProgramLink:
                parsed = hrefs.der_program_parse(der.CurrentDERProgramLink.href)
                program = DERProgramAdapter.fetch_at(parsed.index)
            else:
                program = m.DERProgram()
            
            return Response(dataclass_to_xml(program))
            
            
        
        
    def _admin_derp(self, index: int = -1) -> Response:
        if request.method == 'GET' and index < 0:
            return Response(dataclass_to_xml(DERProgramAdapter.fetch_list()))
        elif request.method == 'GET':
            return Response(dataclass_to_xml(DERProgramAdapter.fetch_edev_all()[index]))
        
        if request.method == 'POST':
            xml = request.data.decode('utf-8')
            data = xml_to_dataclass(request.data.decode('utf-8'))
            
            response = DERProgramAdapter.create(data)
            
            return Response(headers={'Location': response.href}, status=response.statusint)
            
            
        return Response(f"I am {index}, {request.method}")

    def _admin_derp_derc(self, index: int) -> Response:
        if request.method == "POST":
            xml = request.data.decode('utf-8')
            data = xml_to_dataclass(request.data.decode('utf-8'))
            
            if isinstance(data, m.DefaultDERControl):
                results = DERProgramAdapter.create_default_der_control(index, data)
                return Response(headers={'Location': results.href}, status=results.statusint)
            elif isinstance(data, m.DERControl):
                results = DERProgramAdapter.create_der_control(index, data)
                return Response(headers={'Location': results.href}, status=results.statusint)
            
        results = DERProgramAdapter.fetch_der_control_list(index)
        return Response(dataclass_to_xml(results))
        

    def _admin(self) -> Response:
        arg_path = request.args.get('path')
        device = request.args.get('device')
        
        if arg_path == '/enddevices':
            return Response(dataclass_to_xml(EndDeviceAdapter.fetch_list()))
        
        if arg_path.startswith('/edev'):
            edev_path = hrefs.edev_parse(arg_path)
            if edev_path.der_index == hrefs.NO_INDEX:
                return Response(dataclass_to_xml(DERAdapter.fetch_list(edev_index=edev_path.edev_index)))
            elif edev_path.der_sub is None:
                return Response(dataclass_to_xml(DERAdapter.fetch_at(edev_index=edev_path.edev_index, der_index=edev_path.der_index)))
            elif edev_path.der_sub == hrefs.DERSubType.CurrentProgram.value:
                return Response(dataclass_to_xml(DERAdapter.fetch_current_program_at(edev_index=edev_path.edev_index, der_index=edev_path.der_index)))
            
        elif arg_path.startswith("/fsa"):
            
            fsa_path = hrefs.fsa_parse(arg_path)
            
            if fsa_path.fsa_index == hrefs.NO_INDEX:
                return Response(dataclass_to_xml(FSAAdapter.fetch_list()))
            elif fsa_path.fsa_index != hrefs.NO_INDEX and fsa_path.fsa_sub is None:
                return Response(dataclass_to_xml(FSAAdapter.fetch_at(fsa_path.fsa_index)))
            else:
                return Response(dataclass_to_xml(FSAAdapter.fetch_program_list(fsa_path.fsa_index)))
            
                
        
        
        

    def _admin_enddevices(self, index:int = None) -> Response:
        
        return Response(dataclass_to_xml(EndDeviceAdapter.fetch_list()))
    
    def _lfdi_lists(self) -> Response:
        items = []

        for k, v in self.end_devices.__all_end_devices__.items():
            items.append({"key": k, "lfdi": int(v.end_device.lFDI)})

        return Response(json.dumps(items))

    def _admin_edev_fsa(self, edevid: int) -> Response:
        #edev = self.end_devices.get(edevid)
        return Response(json.dumps(json.dumps(self.end_devices.get_fsa_list(edevid=edevid))))

    def _program_lists(self) -> str:
        return render_template("admin/program-lists.html",
                               program_lists=self.server_config.program_lists)

    def _admin_enddevice_list(self) -> str:
        return render_template("admin/end-device-list.html",
                               end_device_list=self.end_devices.get_end_devices())



#import sys
from dataclasses import dataclass, field
import time
from argparse import ArgumentParser
#import asyncio
from pathlib import Path
# from typing import List, Optional, Dict
# from threading import Thread
# from threading import Timer

# import pandas as pd
# import matplotlib.pyplot as plt
#import subprocess
import pvlib
import math
import logging

from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.client import IEEE2030_5_Client
from ieee_2030_5.models import MirrorUsagePoint, MirrorMeterReading, ReadingType, Reading, DERCurveList, DERProgramList
# from ieee_2030_5.utils import serialize_dataclass


def run_inverter(client: IEEE2030_5_Client, capabilities_url: str = "/dcap"):
    print(f"running inverter for {client.hostname}")
    cap = client.device_capability(capabilities_url)
    mups = client.mirror_usage_point_list(url=cap.MirrorUsagePointListLink.href)

    if not mups.results:
        mup_uuid = client.new_uuid()
        # f"p_ac = {p_ac}, s_ac = {s_ac}, q_ac= {q_ac}, PF = {PF}, v_ac = {v_ac}, i_ac = {i_ac}"
        mup_readings = [MirrorMeterReading("p_ac"),
                        MirrorMeterReading("s_ac"),
                        MirrorMeterReading("q_ac"),
                        MirrorMeterReading("PF"),
                        MirrorMeterReading("v_ac"),
                        MirrorMeterReading("i_ac")]
        mup = MirrorUsagePoint(description=f"Inverter {client.hostname}",
                               status=1,
                               MirrorMeterReading=mup_readings)
        resp = client.create_mirror_usage_point(cap.MirrorUsagePointListLink.href,
                                                mup)

        # print(resp)
        status, location = resp
        resp = client.request(location)
        # print(resp)

    # PV module
    sandia_modules = pvlib.pvsystem.retrieve_sam('SandiaMod')
    module = sandia_modules['Canadian_Solar_CS5P_220M___2009_']
    # Inverter model
    sapm_inverters = pvlib.pvsystem.retrieve_sam('cecinverter')
    inverter = sapm_inverters['ABB__MICRO_0_25_I_OUTD_US_208__208V_']
    irradiance = [900, 1000, 925]
    temperature = [25, 28, 20]
    # Assumed constant power factor
    PF = 0.99
    # Assumed constant AC voltage
    v_ac = 120
    latitude = 32
    longitude = -111.0
    weather = pvlib.iotools.get_pvgis_tmy(latitude, longitude, map_variables=True)[0]
    total_solar_radiance = weather['ghi']
    # assumed that the total solar radiance is equal to ghi(global horizontal irradiance)
    outdoor_temp = weather["temp_air"]
    # both the total_solar_radiace and outdoor_temp has 1hr sampling rate.
    # you should be able modify the sampling rate by resampling it
    for x, y in zip(total_solar_radiance, outdoor_temp):
        dc = pvlib.pvsystem.sapm(x, y, module)
        # print(dc)
        p_ac = pvlib.inverter.sandia(dc['v_mp'], dc['p_mp'], inverter)
        s_ac = p_ac / PF
        q_ac = math.sqrt(p_ac**2 + s_ac**2)
        i_ac = (s_ac / v_ac) * 1000

        print(f"p_ac = {p_ac}, s_ac = {s_ac}, q_ac= {q_ac}, PF = {PF}, v_ac = {v_ac}, i_ac = {i_ac}")
        yield p_ac, s_ac, PF, v_ac, i_ac
        # single phase circuit calculation


# def main(clients: List[IEEE2030_5_Client], params: Optional[Dict[int, str]] = None):
#     if not isinstance(clients, list):
#         raise ValueError("clients must be a list")
#
#     if params is None:
#         params = {}
#
#     for index, c in enumerate(clients):
#         if params.get(index):
#             print(f"Got inverter params for index {index} {params.get(index)}")
#         else:
#             print(f"No params for {index}")
#         p_ac, s_ac, PF, v_ac, i_ac = run_inverter(c)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    parser = ArgumentParser()
    parser.add_argument("--tls-repo", help="TLS repository directory to use, defaults to ~/tls", default="~/tls")
    parser.add_argument("--server-host", help="Reference to the utilities server for data.")
    parser.add_argument("--server-port", type=int, default=443,
                        help="Port the server is listening on, default to 443")
    parser.add_argument("--device-id", help="The id of the device for this inverter")
    parser.add_argument("--pin", type=int,
                        help="PIN for the client to validate that it is connecting to the correct server.")

    opts = parser.parse_args()

    path = Path(__file__).parent.parent.parent.joinpath("openssl.cnf")
    repo_dir = Path(opts.tls_repo).expanduser().resolve(strict=True)
    if not repo_dir.exists():
        raise ValueError(f"Invalid repo directory {str(repo_dir)}")
    tls_repo = TLSRepository(repo_dir=repo_dir, openssl_cnffile_template=path,
                             serverhost="gridappsd_dev_2004", clear=False)

    if opts.device_id not in tls_repo.client_list:
        raise ValueError(f"device_id: ({opts.device_id}) not in tls repository")

    # print(new_tls_repository.client_list)
    for p in tls_repo.client_list:
        if p == opts.device_id:
            IEEE2030_5_Client(cafile=tls_repo.ca_cert_file,
                              keyfile=tls_repo.__get_key_file__(opts.device_id),
                              certfile=tls_repo.__get_cert_file__(opts.device_id),
                              hostname=p,
                              server_hostname=opts.server_host,
                              server_ssl_port=opts.server_port)

    # Start up a client from the available in the config file.
    client = list(IEEE2030_5_Client.clients)[0]
    for i in range(5):
        # Device capability provides links to the other resources of interest.
        dcap = client.device_capability()
        time.sleep(2)
    # Time request for offsets to keep things closely aligned.
    tm = client.time()
    # The device that this class is available for.
    device = client.end_device()
    # print("EndDevice")
    # print(serialize_dataclass(device))
    #

    # # Registration to test if we have the correct pin for the client to be sure
    # # that it is talking to the correct server
    # reg = client.registration(device)
    # assert reg.pIN == opts.pin
    #
    # curve_list: DERCurveList = client.__get_request__("/curves")
    # curve1 = client.__get_request__(curve_list.DERCurve[0].href)
    #
    # global_programs: DERProgramList = client.__get_request__("/programs")
    # program = client.__get_request__(global_programs.DERProgram[0].href)
    #
    #
    #
    # der_programs = client.der_program_list(device)
    #
    # der_programs = client.der_program_list()

    #print(fsa)
    # edevs = client.end_devices()
    #
    # my_device = client.end_device()
    # self_device = client.self_device()
    # assert my_device == self_device
    # end_devices = client.end_devices()
    # end_device = client.end_device(0)
    # registration = client.registration(end_device)
    #
    # # edev_config = client.request(end_device.ConfigurationLink.href)
    # client.timelink()
    #
    # assert registration.pIN == opts.pin
    #
    # der_list = client.__get_request__(end_device.DERListLink.href)
    # # uuidstr = client.new_uuid()
    # mup = client.mirror_usage_point_list()
    #
    # mup_uuid = client.new_uuid().encode('utf-8')
    # mup_gas_mirroring = MirrorUsagePoint(
    #     mRID=mup_uuid,
    #     description="Gas Mirroring",
    #     roleFlags=bytes(13),
    #     serviceCategoryKind=1,
    #     status=1,
    #     deviceLFDI=end_device.lFDI,
    #     MirrorMeterReading=[MirrorMeterReading(
    #         mRID=mup_uuid,
    #         Reading=Reading(
    #             value=125
    #         ),
    #         ReadingType=ReadingType(
    #             accumulationBehaviour=9,
    #             commodity=7,
    #             dataQualifier=0,
    #             flowDirection=1,
    #             powerOfTenMultiplier=3,
    #             uom=119
    #         )
    #     )]
    # )
    #
    # status, location = client.create_mirror_usage_point(mup_gas_mirroring)
    #
    # point_list = client.mirror_usage_point_list()
    #
    # print(point_list)

    # print(client.new_uuid())
    # print(client.usage_point_list())
    # gen = run_inverter(client)
    # for output in gen:
    #     print(output)


    # threads: List[Thread] = []
    #
    # for index, client in enumerate(IEEE2030_5_Client.clients):
    #     th = Thread(target=run_inverter, args=(client,))
    #     threads.append(th)
    #     th.daemon = True
    #     th.start()
    #
    # while True:
    #     alive = False
    #     for t in threads:
    #         if t.is_alive():
    #             alive = True
    #             break
    #     if alive:
    #         break
    #     time.sleep(1)


    # while True:
    #     try:
    #         time.sleep(0.1)
    #     except KeyboardInterrupt:
    #         sys.stderr.write("Exiting program\n")
    #         sys.exit(0)



import uuid
from ast import Dict
from datetime import datetime
from typing import List

import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.adapters import AdapterListProtocol, BaseAdapter, ready_signal
from ieee_2030_5.adapters.der import DERAdapter, DERProgramAdapter
from ieee_2030_5.adapters.fsa import FSAAdapter
from ieee_2030_5.data.indexer import add_href
from ieee_2030_5.models.enums import DeviceCategoryType
from ieee_2030_5.types_ import Lfdi


class _EndDeviceAdapter(BaseAdapter, AdapterListProtocol):
    
    def __init__(self) -> None:
        self._end_devices: List[m.EndDevice] = []
        self._reg: Dict[int, m.Registration] = {}
        self._fsa: List[m.FunctionSetAssignments] = []
        self._edev_fsa: Dict[int, List[m.FunctionSetAssignments]] = {}
            
    def fetch_registration(self, edev_index: int) -> m.Registration:
        return self._reg[edev_index]
    
    def fetch_fsa_list(self, edev_index: int, start: int = 0, after: int = 0, limit: int = 0) -> m.FunctionSetAssignmentsList:
        fsa_list = m.FunctionSetAssignmentsList(href=hrefs.fsa_href(edev_index=edev_index), FunctionSetAssignments=self._fsa)
        return fsa_list
    
    def fetch_fsa(self, edev_index: int, fsa_index: int) -> m.FunctionSetAssignments:
        return self._edev_fsa[edev_index][fsa_index]
        
    def __initialize__(self, sender):
        """ Intializes the following based upon the device configuration and the tlsrepository.
        
        Each EndDevice will have the following sub-components initialized:
        - PowerStatus - PowerStatusLink
        - DeviceStatus - DeviceStatusLink
        - Registration - RegistrationLink
        - MessagingProgramList - MessagingProgramListLink
        - Log
        Either FSA or DemandResponseProgram
        - DemandResponseProgram - DemandResponseProgramListLink
        
        
        As well as the following properties
        - changedTime - Current time of initialization
        - sFDI - The short form of the certificate for the system.
        """
        # assert EndDeviceAdapter.__tls_repository__ is not None
        # EndDeviceAdapter.initialize_from_storage()
        # programs = DERProgramAdapter.get_all()
        # stored_devices = EndDeviceAdapter.get_all()
        programs = DERProgramAdapter.fetch_edev_all()

        for dev in BaseAdapter.device_configs():
            edev = m.EndDevice()
            edev.lFDI = BaseAdapter.__tls_repository__.lfdi(dev.id)
            edev.sFDI = BaseAdapter.__tls_repository__.sfdi(dev.id)
            # TODO handle enum eval in a better way.
            edev.deviceCategory = eval(f"DeviceCategoryType.{dev.deviceCategory}")
            edev.enabled = dev.enabled

            # TODO remove subscribable
            edev.subscribable = None
            
            # Add the end device to the list.
            index = self.add(edev)
            
            ts = int(round(datetime.utcnow().timestamp()))
            self._reg[index] = m.Registration(href=hrefs.registration_href(index),
                                              pIN=dev.pin, 
                                              dateTimeRegistered=ts)
            edev.RegistrationLink = m.RegistrationLink(href=hrefs.registration_href(index))
            #edev.FunctionSetAssignmentsListLink = m.FunctionSetAssignmentsListLink(href=hrefs.fsa_href(edev_index=index))
            
            fsa_programs = []
            for cfg_program in dev.programs:
                for program in programs:
                    program.mRID = str(uuid.uuid4())
                    if cfg_program["description"] == program.description:
                        fsa_programs.append(program)
                        break
                    
            if len(fsa_programs) > 0:
                fsa = FSAAdapter.create(fsa_programs)
                edev.FunctionSetAssignmentsListLink = m.FunctionSetAssignmentsListLink(href=hrefs.fsa_href(edev_index=index))
                self._fsa.append(fsa)
                
                if not self._edev_fsa.get(index):
                    self._edev_fsa[index] = [fsa]
                else:
                    self._edev_fsa[index].append(fsa)
            
            has_der = False
            for der_indx, der in enumerate(dev.ders):                
                DERAdapter.create(edev_index=index, der_index=der_indx, modesSupported=der["modesSupported"], deviceType=der["type"])
                has_der = True
            if has_der:
                edev.DERListLink = m.DERListLink(hrefs.edev_der_href(edev_index=index))  # hrefs.der_sub_href(index=index))
                
            #self._end_devices.append(edev)
                            

            # log = m.LogEventList(href=hrefs.get_log_list_href(index),
            #                      all=0,
            #                      results=0,
            #                      pollRate=BaseAdapter.server_config().log_event_list_poll_rate)
            # edev.LogEventListLink = m.LogEventListLink(href=log.href)
            # add_href(log.href, log)

            # cfg = m.Configuration(href=hrefs.get_configuration_href(index))
            # add_href(cfg.href, cfg)
            # edev.ConfigurationLink = m.ConfigurationLink(cfg.href)

            # ps = m.PowerStatus(href=hrefs.get_power_status_href(index))
            # add_href(ps.href, ps)
            # edev.PowerStatusLink = m.PowerStatusLink(href=ps.href)

            # ds = m.DeviceStatus(href=hrefs.get_device_status(index))
            # add_href(ds.href, ds)
            # edev.DeviceStatusLink = m.DeviceStatusLink(href=ds.href)

            # di = m.DeviceInformation(href=hrefs.get_device_information(index))
            # add_href(di.href, di)
            # edev.DeviceInformationLink = m.DeviceInformationLink(href=di.href)

            # ts = int(round(datetime.utcnow().timestamp()))
            # reg = m.Registration(href=hrefs.get_registration_href(index),
            #                      pIN=dev.pin,
            #                      dateTimeRegistered=ts)
            # add_href(reg.href, reg)
            # edev.RegistrationLink = m.RegistrationLink(reg.href)

            # log = m.LogEventList(href=hrefs.get_log_list_href(index), all=0)
            # add_href(log.href, log)
            # edev.LogEventListLink = m.LogEventListLink(log.href)

            # fsa_list = m.FunctionSetAssignmentsList(href=hrefs.get_fsa_list_href(edev.href))

            # fsa = m.FunctionSetAssignments(href=hrefs.get_fsa_href(fsa_list_href=fsa_list.href,
            #                                                        index=0),
            #                                mRID="0F")
            # edev.FunctionSetAssignmentsListLink = m.FunctionSetAssignmentsListLink(fsa_list.href)

            # der_program_list = m.DERProgramList(href=hrefs.get_der_program_list(fsa_href=fsa.href),
            #                                     all=0,
            #                                     results=0)

            # fsa.DERProgramListLink = m.DERProgramListLink(href=der_program_list.href)
            # fsa_list.FunctionSetAssignments.append(fsa)

            # for cfg_program in dev.programs:
            #     for program in programs:
            #         program.mRID = "1F"
            #         if cfg_program["description"] == program.description:
            #             der_program_list.all += 1
            #             der_program_list.results += 1
            #             der_program_list.DERProgram.append(program)
            #             break

            # # Allow der list here
            # # # TODO: instantiate from config file.
            # der_list = m.DERList(
            #     href=hrefs.get_der_list_href(index),
            # #pollRate=900,
            #     results=0,
            #     all=0)
            # edev.DERListLink = m.DERListLink(der_list.href)

            # self._end_devices.append(edev)

            # edev_list.EndDevice.append(edev)
        
    def fetch_edev_all(self) -> List:
        return self._end_devices
    
    def fetch_list(self, start: int = 0, after: int = 0, limit: int = 0) -> m.EndDeviceList:
        enddevice_list = m.EndDeviceList(href=hrefs.get_enddevice_list_href(),
                                         EndDevice=self._end_devices,
                                         all = len(self._end_devices), results=len(self._end_devices))
        return enddevice_list
    
    def get_at(self, index: int) -> m.EndDevice:
        return self._end_devices[index]
    
    def add(self, end_device: m.EndDevice) -> int:
        """Add and return the index of the newly added end device.
        
        This method will also update the href to be the current location of the resource
        """
        end_device.href = hrefs.get_enddevice_href(len(self._end_devices))
        self._end_devices.append(end_device)
        return len(self._end_devices) - 1
    
    def fetch_by_lfdi(self, lfdi: Lfdi) -> m.EndDevice:
        try:
            ed =  next(filter(lambda x: x.lFDI == lfdi, self._end_devices))
            
        except StopIteration:
            ed = None
        
        return ed
    
    def fetch_list_by_lfdi(self, lfdi: Lfdi) -> m.EndDeviceList:
        try:
            ed =  next(filter(lambda x: x.lFDI == lfdi, self._end_devices))
            edl = m.EndDeviceList(href=hrefs.get_enddevice_list_href(),
                            EndDevice=[ed],
                            all=1, results=1)
        except StopIteration:
            edl = m.EndDeviceList(all=0, results=0)
        
        return edl
    
    @staticmethod
    def get_list(lfdi: Lfdi, s: int = 0, l: int = 1) -> m.EndDeviceList:
        ed_list = m.EndDeviceList(href=hrefs.get_enddevice_list_href(), all=0, results=0)

        # TODO remove as we test oeg_client.
        ed_list.pollRate = None
        ed_list.subscribable = None

        for ed in EndDeviceAdapter.get_all():
            if ed.lFDI == lfdi:
                ed_list.all += 1
                ed_list.results += 1
                ed_list.EndDevice.append(ed)

        return ed_list

    @staticmethod
    def initialize_from_storage():
        hrefs_found = get_href_filtered(hrefs.get_enddevice_href(hrefs.NO_INDEX))
        EndDeviceAdapter.__count__ = len(hrefs_found)

    @staticmethod
    def build(**kwargs) -> m.EndDevice:
        ed = m.EndDevice()
        populate_from_kwargs(ed, **kwargs)
        return ed

    @staticmethod
    def find_index(end_device: m.EndDevice) -> int:
        for i in range(EndDeviceAdapter.__count__ + 1):
            if end_device.href == hrefs.get_enddevice_href(i):
                return i

        raise KeyError(f"End device not found for {end_device.href}")

    @staticmethod
    def get_by_index(index: int) -> m.EndDevice:
        return get_href(hrefs.get_enddevice_href(index))

    @staticmethod
    def get_next_href() -> str:
        return hrefs.get_enddevice_href(EndDeviceAdapter.get_next_index())


    @staticmethod
    def get_by_lfdi(lfdi: Lfdi) -> m.EndDevice:
        for ed in EndDeviceAdapter.get_all():
            if ed.lFDI == lfdi:
                return ed
        return None

    @staticmethod
    def store(device_id: str, value: m.EndDevice) -> m.EndDevice:
        """Store the end device into temporary/permanant storage.
        
        The device_id is necessary to map the configured device into the linked registration
        
        This function will add the href and registration link to the end device.  
        
        """
        if not value.href:
            value.href = EndDeviceAdapter.get_next_href()
        if not value.RegistrationLink:
            reg_time = datetime.now(timezone.utc)
            pin = None
            for dev in BaseAdapter.__device_configurations__:
                if dev.id == device_id:
                    pin = dev.pin
                    break

            mreg = m.Registration(href=hrefs.get_registration_href(
                EndDeviceAdapter.find_index(value)),
                                  pIN=pin,
                                  dateTimeRegistered=format_time(reg_time))
            add_href(mreg.href, mreg)
            value.RegistrationLink = m.RegistrationLink(mreg.href)

        add_href(value.href, value)
        return value

    @staticmethod
    def get_all() -> List[m.EndDevice]:
        end_devices: List[m.EndDevice] = []
        href_prefix = hrefs.get_enddevice_href(hrefs.NO_INDEX)
        cpl = re.compile(f"{href_prefix}{hrefs.SEP}[0-9]+$")
        for ed in get_href_filtered(href_prefix=href_prefix):
            if cpl.match(ed.href):
                end_devices.append(ed)

        return sorted(end_devices, key=lambda k: k.href)

EndDeviceAdapter = _EndDeviceAdapter()
ready_signal.connect(EndDeviceAdapter.__initialize__, DERProgramAdapter)




from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from functools import lru_cache
from typing import List, NamedTuple, Optional

EDEV = "edev"
DCAP = "dcap"
UTP = "upt"
MUP = "mup"
DRP = "drp"
SDEV = "sdev"
MSG = "msg"
DER = "der"
CURVE = "dc"
RSPS = "rsps"
LOG = "log"
DERC = "derc"
DDERC = "dderc"
FSA = "fsa"
DERP = "derp"
DERCA = "derca"

DEFAULT_DCAP_ROOT = f"/{DCAP}"
DEFAULT_EDEV_ROOT = f"/{EDEV}"
DEFAULT_UPT_ROOT = f"/{UTP}"
DEFAULT_MUP_ROOT = f"/{MUP}"
DEFAULT_DRP_ROOT = f"/{DRP}"
DEFAULT_SELF_ROOT = f"/{SDEV}"
DEFAULT_MESSAGE_ROOT = f"/{MSG}"
DEFAULT_DER_ROOT = f"/{DER}"
DEFAULT_CURVE_ROOT = f"/{CURVE}"
DEFAULT_RSPS_ROOT = f"/{RSPS}"
DEFAULT_LOG_EVENT_ROOT = f"/{LOG}"
DEFAULT_FSA_ROOT = f"/{FSA}"
DEFAULT_DERP_ROOT = f"/{DERP}"

SEP = "_"
MATCH_REG = "[a-zA-Z0-9_]*"

# Used as a sentinal value when we only want the href of the root
NO_INDEX = -1

class DERSubType(Enum):
    Capability = "dercap"
    Settings = "derg"
    Status = "ders"
    Availability = "dera"
    CurrentProgram = "derp"

class FSASubType(Enum):
    DERProgram = "derp"
    
class DERProgramSubType(Enum):
    NoLink = 0
    ActiveDERControlListLink = 1
    DefaultDERControlLink = 2
    DERControlListLink = 3
    DERCurveListLink= 4
    DERControlReplyTo = 5
    
class DERProgramHref(NamedTuple):
    root: str
    index: int
    
    @staticmethod
    def parse(href: str):
        parsed = href.split(SEP)
        if len(parsed) == 1:
            return DERProgramHref(parsed[0], NO_INDEX)
        elif len(parsed) == 2:
            return DERProgramHref(parsed[0], int(parsed[1]))
    
def der_program_parse(href: str) -> DERProgramHref:
    return DERProgramHref.parse(href)
    
def der_program_href(index: int = NO_INDEX, sub: DERProgramSubType = DERProgramSubType.NoLink, subindex: int = NO_INDEX) -> str:
    if index == NO_INDEX:
        return DEFAULT_DERP_ROOT
    
    if sub == DERProgramSubType.NoLink:
        return SEP.join([DEFAULT_DERP_ROOT, str(index)])
    
    if sub == DERProgramSubType.ActiveDERControlListLink:
        if subindex == NO_INDEX:
            return SEP.join([DEFAULT_DERP_ROOT, str(index), DERCA])
        else:
            return SEP.join([DEFAULT_DERP_ROOT, str(index), DERCA, str(subindex)])
    
    if sub == DERProgramSubType.DefaultDERControlLink:
        if subindex == NO_INDEX:
            return SEP.join([DEFAULT_DERP_ROOT, str(index), DDERC])
        else:
            return SEP.join([DEFAULT_DERP_ROOT, str(index), DDERC, str(subindex)])

    if sub == DERProgramSubType.DERCurveListLink:
        if subindex == NO_INDEX:
            return SEP.join([DEFAULT_DERP_ROOT, str(index),  CURVE])
        else:
            return SEP.join([DEFAULT_DERP_ROOT, str(index), CURVE, str(subindex)])
        
    if sub == DERProgramSubType.DERControlListLink:
        if subindex == NO_INDEX:
            return SEP.join([DEFAULT_DERP_ROOT, str(index),  DERC])
        else:
            return SEP.join([DEFAULT_DERP_ROOT, str(index), DERC, str(subindex)])
        
    if sub == DERProgramSubType.DERControlReplyTo:
        return DEFAULT_RSPS_ROOT
        

@lru_cache()
def get_server_config_href() -> str:
    return "/server/cfg"


@lru_cache()
def get_enddevice_list_href() -> str:
    return DEFAULT_EDEV_ROOT

@lru_cache()
def curve_href(index: int = NO_INDEX) -> str:
    if index == NO_INDEX:
        return DEFAULT_CURVE_ROOT
    
    return SEP.join([DEFAULT_CURVE_ROOT, str(index)])

@lru_cache()
def fsa_href(index: int = NO_INDEX, edev_index: int=NO_INDEX):
    if index == NO_INDEX and edev_index == NO_INDEX:
        return DEFAULT_FSA_ROOT
    elif index != NO_INDEX and edev_index == NO_INDEX:
        return SEP.join([DEFAULT_FSA_ROOT, str(index)])
    elif index == NO_INDEX and edev_index != NO_INDEX:
        return SEP.join([DEFAULT_EDEV_ROOT, str(edev_index), FSA])
    else:
        return SEP.join([DEFAULT_EDEV_ROOT, str(edev_index), FSA, str(index)])
    
def der_href(index: int = NO_INDEX, fsa_index: int = NO_INDEX, edev_index: int = NO_INDEX):
    if index == NO_INDEX and fsa_index == NO_INDEX and edev_index == NO_INDEX:
        return DEFAULT_DER_ROOT
    elif index != NO_INDEX and fsa_index == NO_INDEX and edev_index == NO_INDEX:
        return SEP.join([DEFAULT_DER_ROOT, str(index)])
    elif index == NO_INDEX and fsa_index != NO_INDEX and edev_index == NO_INDEX:
        return SEP.join([DEFAULT_FSA_ROOT, str(fsa_index), DERP])
    elif edev_index != NO_INDEX and fsa_index == NO_INDEX and index == NO_INDEX:
        return SEP.join([DEFAULT_EDEV_ROOT, int(edev_index), FSA])
    elif edev_index != NO_INDEX and fsa_index != NO_INDEX and index == NO_INDEX:
        return SEP.join([DEFAULT_EDEV_ROOT, int(edev_index), FSA, int(fsa_index)])
    else:
        raise ValueError(f"index={index}, fsa_index={fsa_index}, edev_index={edev_index}")
    
def edev_der_href(edev_index: int, der_index: int = NO_INDEX) -> str:
    if der_index == NO_INDEX:
        return SEP.join([DEFAULT_EDEV_ROOT, str(edev_index), DER])
    return SEP.join([DEFAULT_EDEV_ROOT, str(edev_index), DER, str(der_index)])

class EdevHref(NamedTuple):
    edev_index: int
    der_index: int = NO_INDEX
    der_sub: DERSubType = None
    
def edev_parse(path: str) -> EdevHref:
    split_pth = path.split(SEP)
    
    if len(split_pth) == 1:
        return EdevHref(NO_INDEX)
    elif len(split_pth) == 2:
        return EdevHref(int(split_pth[1]))
    elif len(split_pth) == 3:
        return EdevHref(int(split_pth[1]))
    elif len(split_pth) == 4:
        return EdevHref(int(split_pth[1]), int(split_pth[3]))
    elif len(split_pth) == 5:
        return EdevHref(int(split_pth[1]), int(split_pth[3]), split_pth[4])

class FSAHref(NamedTuple):
    fsa_index: NO_INDEX
    fsa_sub: FSASubType = None
    
def fsa_parse(path: str) -> FSAHref:
    split_pth = path.split(SEP)
    
    if len(split_pth) == 1:
        return FSAHref(NO_INDEX)
    elif len(split_pth) == 2:
        return FSAHref(int(split_pth[1]))
    elif len(split_pth) == 3:
        return FSAHref(int(split_pth[1]), fsa_sub=split_pth[2])
    
    raise ValueError("Invalid parsing path.")
    
def der_sub_href(edev_index: int, index: int = NO_INDEX, subtype: DERSubType = None):
    if subtype is None and index == NO_INDEX:
        return SEP.join([DEFAULT_EDEV_ROOT, str(edev_index), DER])
    elif subtype is None:
        return SEP.join([DEFAULT_EDEV_ROOT, str(edev_index), DER, str(index)])
    else:
        return SEP.join([DEFAULT_EDEV_ROOT, str(edev_index), DER, str(index), subtype.value])
    
@lru_cache()    
def mirror_usage_point_href(mirror_usage_point_index: int = NO_INDEX):
    """Mirror Usage Point hrefs
    
       /mup
       /mup/{mirror_usage_point_index}
       
    
    """
    if mirror_usage_point_index == NO_INDEX:
        ret = DEFAULT_MUP_ROOT
    else:
        ret = SEP.join([DEFAULT_MUP_ROOT, str(mirror_usage_point_index)])
    
    return ret

@dataclass
class UsagePointHref:
    usage_point_index: int= NO_INDEX
    meter_reading_list_index: int= NO_INDEX
    meter_reading_index: int= NO_INDEX
    reading_set_index: int= NO_INDEX
    reading_index: int= NO_INDEX
    
    @staticmethod
    def parse(href: str) -> UsagePointHref:
        items = href.split(SEP)
        if len(items) == 1:
            return UsagePointHref()
        
        if len(items) == 2:
            return UsagePointHref(usage_point_index = int(items[1]))

@dataclass
class MirrorUsagePointHref:
    mirror_usage_point_index: int = NO_INDEX
    meter_reading_list_index: int= NO_INDEX
    meter_reading_index: int= NO_INDEX
    reading_set_index: int= NO_INDEX
    reading_index: int= NO_INDEX
    
    @staticmethod
    def parse(href: str) -> MirrorUsagePointHref:
        items = href.split(SEP)
        if len(items) == 1:
            return MirrorUsagePointHref()
        
        if len(items) == 2:
            return MirrorUsagePointHref(items[1])
            

def usage_point_href(usage_point_index: int | str = NO_INDEX,
                     meter_reading_list: bool = False,
                     meter_reading_list_index: int = NO_INDEX,
                     meter_reading_index: int = NO_INDEX,
                     meter_reading_type: bool = False,
                     reading_set: bool = False,
                     reading_set_index: int = NO_INDEX,
                     reading_index: int = NO_INDEX):
    """Usage point hrefs 

       /upt
       /upt/{usage_point_index}
       /upt/{usage_point_index}/mr
       /upt/{usage_point_index}/mr/{meter_reading_index}
       /upt/{usage_point_index}/mr/{meter_reading_index}/rt
       /upt/{usage_point_index}/mr/{meter_reading_index}/rs
       /upt/{usage_point_index}/mr/{meter_reading_index}/rs/{reading_set_index}
       /upt/{usage_point_index}/mr/{meter_reading_index}/rs/{reading_set_index}/r
       /upt/{usage_point_index}/mr/{meter_reading_index}/rs/{reading_set_index}/r/{reading_index}
       
       

    """
    if isinstance(usage_point_index, str):
        base_upt = usage_point_index
    else:
        base_upt = DEFAULT_UPT_ROOT
        
    if usage_point_index == NO_INDEX:
        ret = base_upt        
    else:
        if isinstance(usage_point_index, str):
            arr = [base_upt]
        else:
            arr = [DEFAULT_UPT_ROOT, str(usage_point_index)]
            
        if meter_reading_list:
            if meter_reading_list_index == NO_INDEX:
                arr.extend(["mr"])
            else:
                arr.extend(["mr", str(meter_reading_list_index)])
                
                
        
        ret = SEP.join(arr)
    return ret


def get_der_program_list(fsa_href: str) -> str:
    return SEP.join([fsa_href, "der"])


def get_dr_program_list(fsa_href: str) -> str:
    return SEP.join([fsa_href, "dr"])


def get_fsa_list_href(end_device_href: str) -> str:
    return SEP.join([end_device_href, "fsa"])


def get_response_set_href():
    return DEFAULT_RSPS_ROOT


@lru_cache()
def get_der_list_href(index: int) -> str:
    if index == NO_INDEX:
        ret = DEFAULT_DER_ROOT
    else:
        ret = SEP.join([DEFAULT_DER_ROOT, str(index)])
    return ret


@lru_cache()
def get_enddevice_href(edev_indx: int, subref: str = None) -> str:
    if edev_indx == NO_INDEX:
        ret = DEFAULT_EDEV_ROOT
    elif subref:
        ret = SEP.join([DEFAULT_EDEV_ROOT, f"{edev_indx}", f"{subref}"])
    else:
        ret = SEP.join([DEFAULT_EDEV_ROOT, f"{edev_indx}"])
    return ret


@lru_cache()
def registration_href(edev_index: int) -> str:
    return SEP.join([DEFAULT_EDEV_ROOT, str(edev_index), "rg"])


@lru_cache()
def get_configuration_href(edev_index: int) -> str:
    return get_enddevice_href(edev_index, "cfg")


@lru_cache()
def get_power_status_href(edev_index: int) -> str:
    return get_enddevice_href(edev_index, "ps")


@lru_cache()
def get_device_status(edev_index: int) -> str:
    return get_enddevice_href(edev_index, "ds")


@lru_cache()
def get_device_information(edev_index: int) -> str:
    return get_enddevice_href(edev_index, "di")


@lru_cache()
def get_time_href() -> str:
    # return f"{DEFAULT_DCAP_ROOT}{SEP}tm"
    return f"/tm"


@lru_cache()
def get_log_list_href(edev_index: int) -> str:
    return get_enddevice_href(edev_index, "lel")


@lru_cache()
def get_dcap_href() -> str:
    return f"{DEFAULT_DCAP_ROOT}"


def get_dderc_href() -> str:
    return SEP.join([DEFAULT_DER_ROOT, DDERC])


def get_derc_default_href(derp_index: int) -> str:
    return SEP.join([DEFAULT_DER_ROOT, DDERC, f"{derp_index}"])


def get_derc_href(index: int) -> str:
    """Return the DERControl href to the caller

    if NO_INDEX then don't include the index in the result.
    """
    if index == NO_INDEX:
        return SEP.join([DEFAULT_DER_ROOT, DERC])

    return SEP.join([DEFAULT_DER_ROOT, DERC, f"{index}"])


def get_program_href(index: int, subref: str = None):
    """Return the DERProgram href to the caller

    Args:
        index: if NO_INDEX then don't include the index in the result else use the index
        subref: used to specify a subsection in the program.
    """
    if index == NO_INDEX:
        ref = f"{DEFAULT_DERP_ROOT}"
    else:
        if subref is not None:
            ref = f"{DEFAULT_DERP_ROOT}{SEP}{index}{SEP}{subref}"
        else:
            ref = f"{DEFAULT_DERP_ROOT}{SEP}{index}"
    return ref


# TimeLink
# tm: str = f"{DEFAULT_DCAP_ROOT}{SEP}tm"
# # ResponseSetListLink
# rsps: str = f"{DEFAULT_DCAP_ROOT}{SEP}rsps"
# # UsagePointListLink
# upt: str = DEFAULT_UPT_ROOT
#
# DERProgramListLink
# derp: str = "/derp"
#
# # EndDeviceListLink
# edev: str = DEFAULT_EDEV_ROOT
# edev_urls: List = [
#     f"/<regex('{edev}[0-9a-zA-Z\-]*'):path>",
#     # f"{edev}/<path:fullpath>"
#     # ,
#     # f"{edev}/<int:index>",
#     # f"{edev}/<int:index>/<category>"
# ]
#
# # MirrorUsagePointListLink
# mup: str = DEFAULT_MUP_ROOT
# mup_urls: List = [
#     (mup, ('GET', 'POST')),
#     f"{mup}/<int:index>"
# ]
#
# curve: str = "/curves"
# curve_urls: List = [
#     f"{curve}",
#     (f"{curve}/<int:index>", ("GET",))
# ]
#
# program: str = "/programs"
# program_urls: List = [
#     f"{program}",
#     (f"{program}/<int:index>/actderc", ("GET",)),
#     (f"{program}/<int:index>/dc", ("GET",)),
#     (f"{program}/<int:index>/dderc", ("GET",)),
#     (f"{program}/<int:index>/derc", ("GET",)),
# ]
#
# der: str = "/der"
# der_urls: List = [
#     (f"{der}/<int:edev_id>", ('GET', 'POST')),
#     (f"{der}/<int:edev_id>/<int:id>", ('GET', 'PUT', 'DELETE')),
#     (f"{der}/<int:edev_id>/<int:id>/upt", ('GET', 'DELETE')),
#     (f"{der}/<int:edev_id>/<int:id>/derp", ('GET', 'POST')),
#     (f"{der}/<int:edev_id>/<int:id>/cdp", ('GET', 'DELETE')),
#     (f"{der}/<int:edev_id>/<int:id>/derg", ('GET', 'PUT')),
#     (f"{der}/<int:edev_id>/<int:id>/ders", ('GET', 'PUT')),
#     (f"{der}/<int:edev_id>/<int:id>/dera", ('GET', 'PUT')),
#     (f"{der}/<int:edev_id>/<int:id>/dercap", ('GET', 'PUT')),
# ]
#
sdev: str = DEFAULT_SELF_ROOT


def build_der_link(edev_id: Optional[int] = None,
                   id: Optional[int] = None,
                   suffix: Optional[str] = None) -> str:
    if edev_id is None:
        raise ValueError("edev_id must be specified.")
    if id is not None and suffix is not None:
        link = build_link(f"{der}", f"{edev_id}", f"{id}", suffix)
    elif id is not None:
        link = build_link(f"{der}", f"{edev_id}", f"{id}")
    elif suffix is not None:
        link = build_link(f"{der}", f"{edev_id}", suffix)
    else:
        link = build_link(f"{der}", f"{edev_id}")

    return link


def build_edev_registration_link(index: int) -> str:
    return build_link(f"{edev}", index, "reg")


def build_edev_status_link(index: int) -> str:
    return build_link(f"{edev}", index, "ds")


def build_edev_config_link(index: int) -> str:
    return build_link(f"{edev}", index, "cfg")


def build_edev_info_link(index: int) -> str:
    return build_link(f"{edev}", index, "di")


def build_edev_power_status_link(index: int) -> str:
    return build_link(f"{edev}", index, "ps")


def build_edev_fsa_link(index: int, fsa_index: Optional[int] = None) -> str:
    return build_link(f"{edev}", index, "fsa", fsa_index)


# edev_cfg_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/cfg"
# edev_status_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/ds"
# edev_info_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/di"
# edev_power_status_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/ps"
# edev_file_status_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/fs"
# edev_sub_list_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/subl"

#
# # DemandResponseProgramListLink
# drp: str = DEFAULT_DRP_ROOT
# # MessagingProgramListLink
# msg: str = DEFAULT_MESSAGE_ROOT
# # SelfDeviceLink
# sdev: str = DEFAULT_SELF_ROOT
#
# edev_base: str = '/edev'
# edev: List[str] = [
#     DEFAULT_EDEV_ROOT,
#     [f"{DEFAULT_EDEV_ROOT}/<int:index>", ["GET", "POST"]]
# ]
#
#
#
# sdev_di: str = f"{DEFAULT_DCAP_ROOT}/sdev/di"
# sdev_log: str = f"{DEFAULT_DCAP_ROOT}/sdev/log"
#
# mup_fmt: str = f"{DEFAULT_MUP_ROOT}" + "/{index}"
#
# edev_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}"
# reg_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/reg"
# # di_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/di"
# #dstat_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/dstat"
# ps_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/ps"
#
# derp_list_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/derp"
# derp_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/derp/1"
#
# der_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/der/1"
# dera_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/dera/1"
# dercap_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/dercap/1"
# derg_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/derg/1"
# ders_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/ders/1"
#
# derc_list_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/derc"
# derc_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/derc/1"
#
# fsa_list_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/fsa"
# fsa_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/fsa/0"
#
# edev_cfg_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/cfg"
# edev_status_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/ds"
# edev_info_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/di"
# edev_power_status_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/ps"
# edev_file_status_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/fs"
# edev_sub_list_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/subl"

admin: str = "/admin"
uuid_gen: str = "/uuid"


def build_link(base_url: str, *suffix: Optional[str]):
    result = base_url
    if result.endswith("/"):
        result = result[:-1]

    if suffix:
        for p in suffix:
            if p is not None:
                if isinstance(p, str):
                    if p.startswith("/"):
                        result += f"{p}"
                    else:
                        result += f"/{p}"
                else:
                    result += f"/{p}"

    return result


def extend_url(base_url: str, index: Optional[int] = None, suffix: Optional[str] = None):
    result = base_url
    if index is not None:
        result += f"/{index}"
    if suffix:
        result += f"/{suffix}"

    return result



"""
This module handles MirrorUsagePoint and UsagePoint constructs for a server.
"""
from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Dict, List, Optional

from flask import Response, request
from werkzeug.exceptions import BadRequest

import ieee_2030_5.adapters as adpt
import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.data.indexer import get_href
from ieee_2030_5.server.base_request import RequestOp
from ieee_2030_5.server.uuid_handler import UUIDHandler
from ieee_2030_5.utils import dataclass_to_xml, xml_to_dataclass


class Error(Exception):
    pass


@dataclass
class ResponseStatus:
    location: str
    status: str


class UsagePointRequest(RequestOp):
    
    def get(self) -> Response:
        pth_info = request.environ['PATH_INFO']
        try:
            upt = adpt.MirrorUsagePointAdapter.fetch_usage_point_by_href(pth_info)
        except StopIteration:
            return Response(status=404, response="Not Found")
        return self.build_response_from_dataclass(upt)


class MirrorUsagePointRequest(RequestOp):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self) -> Response:
        pth_info = request.path

        if not pth_info.startswith(hrefs.DEFAULT_MUP_ROOT):
            raise ValueError(f"Invalid path for {self.__class__} {request.path}")
        
        mup_href = hrefs.MirrorUsagePointHref.parse(pth_info)
        try:
            if mup_href.mirror_usage_point_index == hrefs.NO_INDEX:
                mup = adpt.MirrorUsagePointAdapter.fetch_mirror_usage_point_list(pth_info)
            else:
                mup = adpt.MirrorUsagePointAdapter.fetch_mirror_usage_by_href(pth_info)
                
            return self.build_response_from_dataclass(mup)
        except StopIteration:
            if pth_info == hrefs.mirror_usage_point_href():
                mupl = adpt.MirrorUsagePointAdapter.fetch_mirror_usage_point_list()
                
                
            return Response("Not Found", status=404)
        

    def post(self) -> Response:
        xml = request.data.decode('utf-8')
        data = xml_to_dataclass(request.data.decode('utf-8'))
        data_type = type(data)
        if data_type not in (m.MirrorUsagePoint, m.MirrorReadingSet, m.MirrorMeterReading):
            raise BadRequest()

        pth_info = request.path
        pths = pth_info.split(hrefs.SEP)
        if len(pths) == 1 and data_type is not m.MirrorUsagePoint:
            # Check to make sure not a new mrid
            raise BadRequest("Must post MirrorUsagePoint to top level only")

        # Creating a new mup
        if data_type == m.MirrorUsagePoint:
            result = adpt.MirrorUsagePointAdapter.create(mup=data)
        else:
            result = adpt.MirrorUsagePointAdapter.create_reading(href=pth_info, data=data)
        
        result = ResponseStatus(status=result[0], location = result[1])

        if isinstance(result, Error):
            return Response(result.args[1], status=500)
        # Note response to the post is different due to added endpoint.
        
        return Response(headers={'Location': result.location}, status=result.status if isinstance(result.status, int) else result.status.value)



from typing import Optional

from flask import Response, request

import ieee_2030_5.adapters as adpt
import ieee_2030_5.hrefs as hrefs
from ieee_2030_5.server.base_request import RequestOp


class DERRequests(RequestOp):
    """
    Class supporting end devices and any of the subordinate calls to it.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self) -> Response:
        
        if not request.path.startswith(hrefs.DEFAULT_DER_ROOT):
            raise ValueError(f"Invalid path for {self.__class__} {request.path}")
        
        pth_split = request.path.split(hrefs.SEP)
        
        if len(pth_split) == 1:
            # TODO Add arguments
            value = adpt.DERAdapter.fetch_list()
        else:
            value = adpt.DERAdapter.fetch_at(int(pth_split[1]))

        return self.build_response_from_dataclass(value)
    

class DERProgramRequests(RequestOp):
    """
    Class supporting end devices and any of the subordinate calls to it.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self) -> Response:
        pth = request.path
        
        if not pth.startswith(hrefs.DEFAULT_DER_ROOT):
            raise ValueError("Invalid path passed to ")
        obj = self.get_path("foo")
        return self.build_response_from_dataclass(obj)
    



from flask import Response

import ieee_2030_5.adapters as adpt
from ieee_2030_5.server.base_request import RequestOp


class Dcap(RequestOp):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self) -> Response:
        # TODO: Test for allowed dcap here.
        # if not self._end_devices.allowed_to_connect(self.lfdi):
        #     raise werkzeug.exceptions.Unauthorized()
        dcap = adpt.DeviceCapabilityAdapter.get_by_lfdi(self.lfdi)

        return self.build_response_from_dataclass(dcap)


import logging
from typing import Optional

import werkzeug.exceptions
from flask import Response, request

import ieee_2030_5.adapters as adpt
import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.adapters.enddevices import EndDeviceAdapter
from ieee_2030_5.adapters.fsa import FSAAdapter
from ieee_2030_5.data.indexer import get_href
from ieee_2030_5.models import Registration
from ieee_2030_5.server.base_request import RequestOp
from ieee_2030_5.types_ import Lfdi
from ieee_2030_5.utils import dataclass_to_xml, xml_to_dataclass

_log = logging.getLogger(__name__)

class EDevRequests(RequestOp):
    """
    Class supporting end devices and any of the subordinate calls to it.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
    def put(self) -> Response:
        parsed = hrefs.edev_parse(request.path)
        
        if parsed.der_sub not in [x.value for x in hrefs.DERSubType]:
            raise ValueError("Invalid subtype specified.")
        
        response_code = adpt.DERAdapter.store(parsed, xml_to_dataclass(request.data.decode('utf-8')))
        return Response(status=int(response_code))                                              
                                              
                        
        
    def post(self, path: Optional[str] = None) -> Response:
        """
        Handle post request to /edev
        
        The expectation is that data will be an xml object like the following:
        
            <EndDevice xmlns="urn:ieee:std:2030.5:ns">
                <sFDI>231589308001</sFDI>
                <changedTime>0</changedTime>
            </EndDevice>
        
        Args:
            path: 

        Returns:

        """
        # request.data should have xml data.
        if not request.data:
            raise werkzeug.exceptions.Forbidden()

        ed: m.EndDevice = xml_to_dataclass(request.data.decode('utf-8'))

        if not isinstance(ed, m.EndDevice):
            raise werkzeug.exceptions.Forbidden()

        # This is what we should be using to get the device id of the registered end device.
        device_id = self.tls_repo.find_device_id_from_sfdi(ed.sFDI)
        ed.lFDI = self.tls_repo.lfdi(device_id)
        if end_device := adpt.EndDeviceAdapter.fetch_by_lfdi(ed.lfdi):
            status = 200
            ed_href = end_device.href
        else:
            if not ed.href:
                ed = adpt.EndDeviceAdapter.store(device_id, ed)

            ed_href = ed.href
            status = 201

        return Response(status=status, headers={'Location': ed_href})

    def get(self) -> Response:
        """
        Supports the get request for end_devices(EDev) and end_device_list_link.

        Paths:
            /edev
            /edev/0
            /edev/0/di
            /edev/0/rg
            /edev/0/der

        """
        pth = request.path

        if not pth.startswith(hrefs.DEFAULT_EDEV_ROOT):
            raise ValueError(f"Invalid path for {self.__class__} {request.path}")

        pth_split = pth.split(hrefs.SEP)
        
        if len(pth_split) == 1:
            retval = adpt.EndDeviceAdapter.fetch_list_by_lfdi(self.lfdi)
        elif len(pth_split) == 3:
            if pth_split[2] == "rg":
                retval = adpt.EndDeviceAdapter.fetch_registration(edev_index=int(pth_split[1]))
            elif pth_split[2] == "di":
                retval = "foo"
            elif pth_split[2] == "fsa":
                retval = adpt.EndDeviceAdapter.fetch_fsa_list(edev_index=int(pth_split[1]))
            elif pth_split[2] == "der":
                retval = adpt.DERAdapter.fetch_list(edev_index=int(pth_split[1]))
            
        return self.build_response_from_dataclass(retval)


class SDevRequests(RequestOp):
    """
    SelfDevice is an alias for the end device of a client.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self) -> Response:
        """
        Supports the get request for end_devices(EDev) and end_device_list_link.

        Paths:
            /sdev

        """
        end_device = self._end_devices.get_end_device_list(self.lfdi).EndDevice[0]
        return self.build_response_from_dataclass(end_device)

class FSARequests(RequestOp):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
    def get(self):
        pth_split = request.path.split(hrefs.SEP)
        
        if len(pth_split) == 1:
            retval = FSAAdapter.fetch_list()
        elif len(pth_split) == 2:
            retval = FSAAdapter.fetch_at(int(pth_split[1]))
        elif len(pth_split) == 3:
            retval = EndDeviceAdapter.fetch_fsa_list(edev_index=int(pth_split[1]))
        elif len(pth_split) == 4:
            retval = EndDeviceAdapter.fetch_fsa(edev_index=int(pth_split[1]), fsa_index=int(pth_split[3]))
        else:
            raise ValueError(f"Path split is {pth_split}")
            
        return self.build_response_from_dataclass(retval)


import logging
from functools import lru_cache
from typing import Dict, List

import ieee_2030_5.hrefs as hrefs
import ieee_2030_5.models as m
from ieee_2030_5.adapters import AdapterListProtocol, BaseAdapter
from ieee_2030_5.types_ import Lfdi

_log = logging.getLogger(__name__)


__all__: List[str] = [
    "FSAAdapter"
]

class _FSAAdapter(BaseAdapter, AdapterListProtocol):
    def __init__(self) -> None:
        super().__init__()
        self._fsa: List[m.FunctionSetAssignments] = []
        # List of programs assigned to the index of a fsa
        self._programs: Dict[int, List[m.DERProgram]] = {}
            
    def fetch_edev_all(self) -> List[m.FunctionSetAssignments]:
        return self._fsa
    
    def fetch_list(self, start: int = 0, after: int = 0, limit: int = 0) -> m.FunctionSetAssignmentsList:
        fsa_list = m.FunctionSetAssignmentsList(href=hrefs.fsa_href(), all=len(self._fsa), results=len(self._fsa),
                                                FunctionSetAssignments=self._fsa)
        
        return fsa_list
    
    def fetch_program_list(self, fsa_index: int, start: int = 0, after: int = 0, limit: int = 0) -> m.DERProgramList:
        all = len(self._programs[fsa_index])
        program_list = m.DERProgramList(href=hrefs.der_href(fsa_index=fsa_index),
                                        DERProgram=self._programs[fsa_index],
                                        all=all,
                                        results=all)
        return program_list
    
    def fetch_at(self, index: int) -> m.FunctionSetAssignments:
        return self._fsa[index]
    
    def add(self, fsa: m.FunctionSetAssignments) -> int:
        index = len(self._fsa)
        fsa.href = hrefs.fsa_href(index=index)
        return index
    
    def create(self, programs: List[m.DERProgram]) -> m.FunctionSetAssignments:
        fsa = m.FunctionSetAssignments(href=hrefs.fsa_href(index=len(self._fsa)))
        index = self.add(fsa)
        fsa.DERProgramListLink = m.DERProgramListLink(href=hrefs.der_href(fsa_index=index))
        self._programs[index] = programs
        return fsa
        
    
    
FSAAdapter = _FSAAdapter()


from dataclasses import dataclass
from pathlib import Path
from typing import Type, Optional

from xsdata.formats.dataclass.context import XmlContext
from xsdata.formats.dataclass.parsers.config import ParserConfig
from xsdata.formats.dataclass.parsers.xml import XmlParser
from xsdata.formats.dataclass.serializers import XmlSerializer
from xsdata.formats.dataclass.serializers.config import SerializerConfig

__xml_context__ = XmlContext()
__parser_config__ = ParserConfig(fail_on_unknown_attributes=False,
                                 fail_on_unknown_properties=False)
__xml_parser__ = XmlParser(config=__parser_config__, context=__xml_context__)
__config__ = SerializerConfig(xml_declaration=False, pretty_print=True)
__serializer__ = XmlSerializer(config=__config__)
__ns_map__ = {None: "urn:ieee:std:2030.5:ns"}

import ieee_2030_5.types_ as t
import ieee_2030_5.utils.tls_wrapper as tls


def serialize_dataclass(obj: dataclass) -> str:
    """
    Serializes a dataclass that was created via xsdata to an xml string for
    returning to a client.
    """
    return __serializer__.render(obj, ns_map=__ns_map__)


def xml_to_dataclass(xml: str, type: Optional[Type] = None) -> dataclass:
    """
    Parse the xml passed and return result from loaded classes.
    """
    return __xml_parser__.from_string(xml, type)


def dataclass_to_xml(dc: dataclass) -> str:
    return serialize_dataclass(dc)


def get_lfdi_from_cert(path: Path) -> t.Lfdi:
    """
    Using the fingerprint of the certifcate return the left truncation of 160 bits with no check digit.
    Example:
      From:
        3E4F-45AB-31ED-FE5B-67E3-43E5-E456-2E31-984E-23E5-349E-2AD7-4567-2ED1-45EE-213A
      Return:
        3E4F-45AB-31ED-FE5B-67E3-43E5-E456-2E31-984E-23E5
        as an integer.
    """

    # 160 / 4 == 40
    fp = tls.OpensslWrapper.tls_get_fingerprint_from_cert(path)
    fp = fp.replace(":", "")
    lfdi = t.Lfdi(fp[:40])
    return lfdi


def get_sfdi_from_lfdi(lfdi: t.Lfdi) -> int:
    """

    Args:
        lfdi:

    Returns:

    """
    from ieee_2030_5.certs import sfdi_from_lfdi
    return sfdi_from_lfdi(lfdi)



import inspect
import logging
import typing
from dataclasses import dataclass, fields, is_dataclass
from enum import Enum
from typing import Any, Dict, List, Optional, Protocol

from blinker import Signal

import ieee_2030_5.config as cfg
import ieee_2030_5.models as m
from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.models.sep import List_type

_log = logging.getLogger(__name__)


class ReturnCode(Enum):
    OK = 200
    CREATED = 201
    NO_CONTENT = 204
    BAD_REQUEST = 400


def populate_from_kwargs(obj: object, **kwargs) -> Dict[str, Any]:

    if not is_dataclass(obj):
        raise ValueError(f"The passed object {obj} is not a dataclass.")

    for k in fields(obj):
        if k.name in kwargs:
            type_eval = eval(k.type)

            if typing.get_args(type_eval) is typing.get_args(Optional[int]):
                setattr(obj, k.name, int(kwargs[k.name]))
            elif typing.get_args(k.type) is typing.get_args(Optional[bool]):
                setattr(obj, k.name, bool(kwargs[k.name]))
            # elif bytes in args:
            #     setattr(obj, k.name, bytes(kwargs[k.name]))
            else:
                setattr(obj, k.name, kwargs[k.name])
            kwargs.pop(k.name)
    return kwargs

class AdapterIndexProtocol(Protocol):
    
    def fetch_at(self, index: int) -> m.Resource:
        pass


class AdapterListProtocol(AdapterIndexProtocol):
    def fetch_list(self, start: int = 0, after: int =0, limit:int = 0) -> m.List_type:
        pass
    
    def fetch_edev_all(self) -> List:
        pass
        

ready_signal = Signal("ready-signal")

class BaseAdapter:
    __count__: int = 0
    __server_configuration__: cfg.ServerConfiguration
    __tls_repository__: cfg.TLSRepository = None
    __lfdi__mapped_configuration__: Dict[str, cfg.DeviceConfiguration] = {}
    after_initialized = Signal('after-initialized')
        
    @classmethod
    def get_next_index(cls) -> int:
        """Retrieve the next index for an adapter list."""
        return cls.__count__

    @classmethod
    def increment_index(cls) -> int:
        """Increment the list to the next index and return the result to the caller.
        
        
        """
        next = cls.get_next_index()
        cls.__count__ += 1
        return next
    
    @classmethod
    def ready(cls) -> Signal:
        return Signal("ready")

    @classmethod
    def get_current_index(cls) -> int:
        return cls.__count__

    @staticmethod
    def server_config() -> cfg.ServerConfiguration:
        return BaseAdapter.__server_configuration__

    @staticmethod
    def device_configs() -> List[cfg.DeviceConfiguration]:
        return BaseAdapter.__server_configuration__.devices

    @staticmethod
    def tls_repo() -> cfg.TLSRepository:
        return BaseAdapter.__tls_repository__

    @staticmethod
    def get_config_from_lfdi(lfdi: str) -> Optional[cfg.DeviceConfiguration]:
        return BaseAdapter.__lfdi__mapped_configuration__.get(lfdi)

    @staticmethod
    def is_initialized():
        return BaseAdapter.__device_configurations__ is not None and BaseAdapter.__tls_repository__ is not None

    @staticmethod
    def initialize(server_config: cfg.ServerConfiguration, tlsrepo: TLSRepository):
        """Initialize all of the adapters
        
        The initialization means that there are concrete object backing the storage system based upon
        urls that can be read during the http call to the spacific end point.  In other words a
        DERCurve dataclass can be retrieved from storage by going to the href /dc/1 rather than
        having to get it through an object.  
        
        The adapters are responsible for storing data into the object store using add_href function.
        """
        BaseAdapter.__server_configuration__ = server_config
        BaseAdapter.__lfdi__mapped_configuration__ = {}
        BaseAdapter.__tls_repository__ = tlsrepo

        # Map from the configuration id and lfdi to the device configuration.
        for cfg in server_config.devices:
            lfdi = tlsrepo.lfdi(cfg.id)
            BaseAdapter.__lfdi__mapped_configuration__[lfdi] = cfg

        #BaseAdapter.after_initialized.send(BaseAdapter)
        ready_signal.send(BaseAdapter)
        # BaseAdapter.ready().send(BaseAdapter)
        # Find subclasses of us and initialize them calling _initalize method
        # TODO make this non static
        #EndDeviceAdapter._initialize()

    @staticmethod
    def build(**kwargs) -> dataclass:
        raise NotImplementedError()

    @staticmethod
    def store(value: dataclass) -> dataclass:
        raise NotImplementedError()

    @staticmethod
    def build_instance(cls, cfg_dict: Dict, signature_cls=None) -> object:
        if signature_cls is None:
            signature_cls = cls
        return cls(**{
            k: v
            for k, v in cfg_dict.items() if k in inspect.signature(signature_cls).parameters
        })

from ieee_2030_5.adapters.dcap import DeviceCapabilityAdapter
from ieee_2030_5.adapters.der import (DERAdapter, DERControlAdapter,
                                      DERCurveAdapter, DERProgramAdapter)
from ieee_2030_5.adapters.enddevices import EndDeviceAdapter
from ieee_2030_5.adapters.log import LogAdapter
from ieee_2030_5.adapters.mupupt import MirrorUsagePointAdapter



import time
from datetime import datetime
from threading import Thread

from blinker import Signal


class _TimeAdapter(Thread):
    tick = Signal("tick")
    def __init__(self):
        super().__init__()
        self._tick = 0
        
    @staticmethod
    def user_readable(timestamp: int):
        dt = datetime.fromtimestamp(timestamp)
        return dt.strftime("%m/%d/%Y, %H:%M:%S")
    
    def run(self) -> None:
        
        while True:
            self._tick = int(time.mktime(datetime.utcnow().timetuple()))
            _TimeAdapter.tick.send(self._tick)
            time.sleep(1)
        
    
TimeAdapter = _TimeAdapter()
TimeAdapter.daemon = True
TimeAdapter.start()



import subprocess
from pathlib import Path


class TLSWrap:

    @staticmethod
    def tls_create_private_key(file_path: Path):
        """
        Creates a private key in the path that is specified.  The path will be overwritten
        if it already exists.

        Args:
            file_path:

        Returns:

        """
        raise NotImplementedError()

    @staticmethod
    def tls_create_ca_certificate(common_name: str, private_key_file: Path, ca_cert_file: Path):
        """
        Create a ca certificate from using common name private key and ca certificate file.

        Args:
            common_name:
            private_key_file:
            ca_cert_file:

        Returns:

        """
        raise NotImplementedError()

    @staticmethod
    def tls_create_csr(common_name: str, private_key_file: Path, server_csr_file: Path):
        """

        Args:
            common_name:
            private_key_file:
            server_csr_file:

        Returns:

        """
        raise NotImplementedError()

    @staticmethod
    def tls_create_signed_certificate(common_name: str,
                                      ca_key_file: Path,
                                      ca_cert_file: Path,
                                      private_key_file: Path,
                                      cert_file: Path,
                                      as_server: bool = False):
        """

        Args:
            common_name:
            ca_key_file:
            ca_cert_file:
            private_key_file:
            cert_file:
            as_server:

        Returns:

        """
        raise NotImplementedError()

    @staticmethod
    def tls_get_fingerprint_from_cert(cert_file: Path, algorithm: str = "sha256"):
        """

        Args:
            cert_file:
            algorithm:

        Returns:

        """
        raise NotImplementedError()

    @staticmethod
    def tls_create_pkcs23_pem_and_cert(private_key_file: Path, cert_file: Path,
                                       combined_file: Path):
        """

        Args:
            private_key_file:
            cert_file:
            combined_file:

        Returns:

        """
        raise NotImplementedError()


class OpensslWrapper(TLSWrap):
    opensslcnf: Path = None

    def __init__(self, opensslconf: Path):
        OpensslWrapper.opensslcnf = opensslconf

    @staticmethod
    def __set_cnf_from_cert_path___(path: Path):
        if OpensslWrapper.opensslcnf is None:
            check_path = path.parent.parent.joinpath("openssl.cnf")
            if check_path.exists():
                OpensslWrapper.opensslcnf = check_path

    @staticmethod
    def tls_create_private_key(file_path: Path):
        OpensslWrapper.__set_cnf_from_cert_path___(file_path)
        # openssl ecparam -out private/ec-cakey.pem -name prime256v1 -genkey
        cmd = ["openssl", "ecparam", "-out", str(file_path), "-name", "prime256v1", "-genkey"]
        return subprocess.check_output(cmd, text=True)

    @staticmethod
    def tls_create_ca_certificate(common_name: str, private_key_file: Path, ca_cert_file: Path):
        OpensslWrapper.__set_cnf_from_cert_path___(ca_cert_file)
        # openssl req -new -x509 -days 3650 -config openssl.cnf \
        #   -extensions v3_ca -key private/ec-cakey.pem -out certs/ec-cacert.pem
        cmd = [
            "openssl", "req", "-new", "-x509", "-days", "3650", "-subj", f"/C=US/CN={common_name}",
            "-config",
            str(OpensslWrapper.opensslcnf), "-extensions", "v3_ca", "-key",
            str(private_key_file), "-out",
            str(ca_cert_file)
        ]
        return subprocess.check_output(cmd, text=True)

    @staticmethod
    def tls_create_csr(common_name: str, private_key_file: Path, server_csr_file: Path):
        OpensslWrapper.__set_cnf_from_cert_path___(private_key_file)
        subject_name = common_name.split(":")[0]
        # openssl req -new -key server.key -out server.csr -sha256
        cmd = [
            "openssl", "req", "-new", "-config",
            str(OpensslWrapper.opensslcnf), "-subj", f"/C=US/CN={subject_name}", "-key",
            str(private_key_file), "-out",
            str(server_csr_file), "-sha256"
        ]
        return subprocess.check_output(cmd, text=True)

    @staticmethod
    def tls_create_signed_certificate(common_name: str,
                                      ca_key_file: Path,
                                      ca_cert_file: Path,
                                      private_key_file: Path,
                                      cert_file: Path,
                                      as_server: bool = False):
        OpensslWrapper.__set_cnf_from_cert_path___(cert_file)
        subject_name = common_name.split(":")[0]
        csr_file = Path(f"/tmp/{common_name}")
        OpensslWrapper.tls_create_csr(common_name, private_key_file, csr_file)
        # openssl ca -keyfile /root/tls/private/ec-cakey.pem -cert /root/tls/certs/ec-cacert.pem \
        #   -in server.csr -out server.crt -config /root/tls/openssl.cnf
        cmd = [
            "openssl",
            "ca",
            "-keyfile",
            str(ca_key_file),
            "-cert",
            str(ca_cert_file),
            "-subj",
            f"/C=US/CN={subject_name}",
            "-in",
            str(csr_file),
            "-out",
            str(cert_file),
            "-config",
            str(OpensslWrapper.opensslcnf),
        # For no prompt use -batch
            "-batch"
        ]
        # if as_server:
        #     "-server"
        #print(" ".join(cmd))
        ret_value = subprocess.check_output(cmd, text=True)
        csr_file.unlink()
        return ret_value

    @staticmethod
    def tls_get_fingerprint_from_cert(cert_file: Path, algorithm: str = "sha256"):
        OpensslWrapper.__set_cnf_from_cert_path___(cert_file)
        if algorithm == "sha256":
            algorithm = "-sha256"
        else:
            raise NotImplementedError()

        cmd = ["openssl", "x509", "-in", str(cert_file), "-noout", "-fingerprint", algorithm]
        ret_value = subprocess.check_output(cmd, text=True)
        if "=" in ret_value:
            ret_value = ret_value.split("=")[1].strip()
        return ret_value

    @staticmethod
    def tls_create_pkcs23_pem_and_cert(private_key_file: Path, cert_file: Path,
                                       combined_file: Path):
        OpensslWrapper.__set_cnf_from_cert_path___(cert_file)
        # openssl pkcs12 -export -in certificate.pem -inkey privatekey.pem -out cert-and-key.pfx
        tmpfile = Path("/tmp/tmp.p12")
        tmpfile2 = Path("/tmp/all.pem")
        tmpfile.unlink(missing_ok=True)
        cmd = [
            "openssl", "pkcs12", "-export", "-in",
            str(cert_file), "-inkey",
            str(private_key_file), "-out",
            str(tmpfile), "-passout", "pass:"
        ]
        subprocess.check_output(cmd, text=True)

        # openssl pkcs12 -in path.p12 -out newfile.pem -nodes
        cmd = [
            "openssl", "pkcs12", "-in",
            str(tmpfile), "-out",
            str(tmpfile2), "-nodes", "-passin", "pass:"
        ]
        # cmd = ["openssl", "pkcs12", "-in", str(tmpfile),
        # "-out", str(combined_file), "-clcerts", "-nokeys", "-passin", "pass:"]
        # -clcerts -nokeys
        subprocess.check_output(cmd, text=True)

        with open(combined_file, "w") as fp:
            in_between = False
            for line in tmpfile2.read_text().split("\n"):
                if not in_between:
                    if "BEGIN" in line:
                        fp.write(f"{line}\n")
                        in_between = True
                else:
                    fp.write(f"{line}\n")
                    if "END" in line:
                        in_between = False



from ieee_2030_5.models.enums import DeviceCategoryType, CurveType 
from ieee_2030_5.models.derforecasts import (
    DERFlexibility,
    DERForecast,
    DERForecastLink,
    ForecastNumericType,
    ForecastParameter,
    ForecastParameterSet,
    ForecastParameterSetList,
)
from ieee_2030_5.models.sep import (
    AbstractDevice,
    AccountBalance,
    AccountBalanceLink,
    AccountingUnit,
    ActiveBillingPeriodListLink,
    ActiveCreditRegisterListLink,
    ActiveDERControlListLink,
    ActiveEndDeviceControlListLink,
    ActiveFlowReservationListLink,
    ActivePower,
    ActiveProjectionReadingListLink,
    ActiveSupplyInterruptionOverrideListLink,
    ActiveTargetReadingListLink,
    ActiveTextMessageListLink,
    ActiveTimeTariffIntervalListLink,
    AmpereHour,
    ApparentPower,
    ApplianceLoadReduction,
    AppliedTargetReduction,
    AssociatedDERProgramListLink,
    AssociatedUsagePointLink,
    BillingMeterReadingBase,
    BillingPeriod,
    BillingPeriodList,
    BillingPeriodListLink,
    BillingReading,
    BillingReadingList,
    BillingReadingListLink,
    BillingReadingSet,
    BillingReadingSetList,
    BillingReadingSetListLink,
    Charge,
    Condition,
    Configuration,
    ConfigurationLink,
    ConnectStatusType,
    ConsumptionTariffInterval,
    ConsumptionTariffIntervalList,
    ConsumptionTariffIntervalListLink,
    CreditRegister,
    CreditRegisterList,
    CreditRegisterListLink,
    CreditTypeChange,
    CurrentDERProgramLink,
    CurrentRMS,
    CurveData,
    CustomerAccount,
    CustomerAccountLink,
    CustomerAccountList,
    CustomerAccountListLink,
    CustomerAgreement,
    CustomerAgreementList,
    CustomerAgreementListLink,
    DER,
    DERAvailability,
    DERAvailabilityLink,
    DERCapability,
    DERCapabilityLink,
    DERControl,
    DERControlBase,
    DERControlList,
    DERControlListLink,
    DERControlResponse,
    DERCurve,
    DERCurveLink,
    DERCurveList,
    DERCurveListLink,
    DERLink,
    DERList,
    DERListLink,
    DERProgram,
    DERProgramLink,
    DERProgramList,
    DERProgramListLink,
    DERSettings,
    DERSettingsLink,
    DERStatus,
    DERStatusLink,
    DRLCCapabilities,
    DateTimeInterval,
    DefaultDERControl,
    DefaultDERControlLink,
    DemandResponseProgram,
    DemandResponseProgramLink,
    DemandResponseProgramList,
    DemandResponseProgramListLink,
    DeviceCapability,
    DeviceCapabilityLink,
    DeviceInformation,
    DeviceInformationLink,
    DeviceStatus,
    DeviceStatusLink,
    DrResponse,
    DutyCycle,
    EndDevice,
    EndDeviceControl,
    EndDeviceControlList,
    EndDeviceControlListLink,
    EndDeviceLink,
    EndDeviceList,
    EndDeviceListLink,
    EnvironmentalCost,
    Error,
    Event,
    EventStatus,
    File,
    FileLink,
    FileList,
    FileListLink,
    FileStatus,
    FileStatusLink,
    FixedPointType,
    FixedVar,
    FlowReservationRequest,
    FlowReservationRequestList,
    FlowReservationRequestListLink,
    FlowReservationResponse,
    FlowReservationResponseList,
    FlowReservationResponseListLink,
    FlowReservationResponseResponse,
    FreqDroopType,
    FunctionSetAssignments,
    FunctionSetAssignmentsBase,
    FunctionSetAssignmentsList,
    FunctionSetAssignmentsListLink,
    GPSLocationType,
    HistoricalReading,
    HistoricalReadingList,
    HistoricalReadingListLink,
    IEEE_802_15_4,
    IPAddr,
    IPAddrList,
    IPAddrListLink,
    IPInterface,
    IPInterfaceList,
    IPInterfaceListLink,
    IdentifiedObject,
    InverterStatusType,
    LLInterface,
    LLInterfaceList,
    LLInterfaceListLink,
    Link,
    List_type,
    ListLink,
    LoadShedAvailability,
    LoadShedAvailabilityList,
    LoadShedAvailabilityListLink,
    LocalControlModeStatusType,
    LogEvent,
    LogEventList,
    LogEventListLink,
    ManufacturerStatusType,
    MessagingProgram,
    MessagingProgramList,
    MessagingProgramListLink,
    MeterReading,
    MeterReadingBase,
    MeterReadingLink,
    MeterReadingList,
    MeterReadingListLink,
    MirrorMeterReading,
    MirrorMeterReadingList,
    MirrorReadingSet,
    MirrorUsagePoint,
    MirrorUsagePointList,
    MirrorUsagePointListLink,
    Neighbor,
    NeighborList,
    NeighborListLink,
    Notification,
    NotificationList,
    NotificationListLink,
    Offset,
    OperationalModeStatusType,
    PEVInfo,
    PowerConfiguration,
    PowerFactor,
    PowerFactorWithExcitation,
    PowerStatus,
    PowerStatusLink,
    PrepayOperationStatus,
    PrepayOperationStatusLink,
    Prepayment,
    PrepaymentLink,
    PrepaymentList,
    PrepaymentListLink,
    PriceResponse,
    PriceResponseCfg,
    PriceResponseCfgList,
    PriceResponseCfgListLink,
    ProjectionReading,
    ProjectionReadingList,
    ProjectionReadingListLink,
    RPLInstance,
    RPLInstanceList,
    RPLInstanceListLink,
    RPLSourceRoutes,
    RPLSourceRoutesList,
    RPLSourceRoutesListLink,
    RandomizableEvent,
    RateComponent,
    RateComponentLink,
    RateComponentList,
    RateComponentListLink,
    ReactivePower,
    ReactiveSusceptance,
    Reading,
    ReadingBase,
    ReadingLink,
    ReadingList,
    ReadingListLink,
    ReadingSet,
    ReadingSetBase,
    ReadingSetList,
    ReadingSetListLink,
    ReadingType,
    ReadingTypeLink,
    RealEnergy,
    Registration,
    RegistrationLink,
    RequestStatus,
    Resource,
    RespondableIdentifiedObject,
    RespondableResource,
    RespondableSubscribableIdentifiedObject,
    Response,
    ResponseList,
    ResponseListLink,
    ResponseSet,
    ResponseSetList,
    ResponseSetListLink,
    SelfDevice,
    SelfDeviceLink,
    ServiceChange,
    ServiceSupplier,
    ServiceSupplierLink,
    ServiceSupplierList,
    SetPoint,
    SignedRealEnergy,
    StateOfChargeStatusType,
    StorageModeStatusType,
    SubscribableIdentifiedObject,
    SubscribableList,
    SubscribableResource,
    Subscription,
    SubscriptionBase,
    SubscriptionList,
    SubscriptionListLink,
    SupplyInterruptionOverride,
    SupplyInterruptionOverrideList,
    SupplyInterruptionOverrideListLink,
    SupportedLocale,
    SupportedLocaleList,
    SupportedLocaleListLink,
    TargetReading,
    TargetReadingList,
    TargetReadingListLink,
    TargetReduction,
    TariffProfile,
    TariffProfileLink,
    TariffProfileList,
    TariffProfileListLink,
    Temperature,
    TextMessage,
    TextMessageList,
    TextMessageListLink,
    TextResponse,
    Time,
    TimeConfiguration,
    TimeLink,
    TimeTariffInterval,
    TimeTariffIntervalList,
    TimeTariffIntervalListLink,
    UnitValueType,
    UnsignedFixedPointType,
    UsagePoint,
    UsagePointBase,
    UsagePointLink,
    UsagePointList,
    UsagePointListLink,
    VoltageRMS,
    WattHour,
    loWPAN,
)

__all__ = [
    "DERFlexibility",
    "DERForecast",
    "DERForecastLink",
    "ForecastNumericType",
    "ForecastParameter",
    "ForecastParameterSet",
    "ForecastParameterSetList",
    "AbstractDevice",
    "AccountBalance",
    "AccountBalanceLink",
    "AccountingUnit",
    "ActiveBillingPeriodListLink",
    "ActiveCreditRegisterListLink",
    "ActiveDERControlListLink",
    "ActiveEndDeviceControlListLink",
    "ActiveFlowReservationListLink",
    "ActivePower",
    "ActiveProjectionReadingListLink",
    "ActiveSupplyInterruptionOverrideListLink",
    "ActiveTargetReadingListLink",
    "ActiveTextMessageListLink",
    "ActiveTimeTariffIntervalListLink",
    "AmpereHour",
    "ApparentPower",
    "ApplianceLoadReduction",
    "AppliedTargetReduction",
    "AssociatedDERProgramListLink",
    "AssociatedUsagePointLink",
    "BillingMeterReadingBase",
    "BillingPeriod",
    "BillingPeriodList",
    "BillingPeriodListLink",
    "BillingReading",
    "BillingReadingList",
    "BillingReadingListLink",
    "BillingReadingSet",
    "BillingReadingSetList",
    "BillingReadingSetListLink",
    "Charge",
    "Condition",
    "Configuration",
    "ConfigurationLink",
    "ConnectStatusType",
    "ConsumptionTariffInterval",
    "ConsumptionTariffIntervalList",
    "ConsumptionTariffIntervalListLink",
    "CreditRegister",
    "CreditRegisterList",
    "CreditRegisterListLink",
    "CreditTypeChange",
    "CurrentDERProgramLink",
    "CurrentRMS",
    "CurveData",
    "CurveType",
    "CustomerAccount",
    "CustomerAccountLink",
    "CustomerAccountList",
    "CustomerAccountListLink",
    "CustomerAgreement",
    "CustomerAgreementList",
    "CustomerAgreementListLink",
    "DER",
    "DERAvailability",
    "DERAvailabilityLink",
    "DERCapability",
    "DERCapabilityLink",
    "DERControl",
    "DERControlBase",
    "DERControlList",
    "DERControlListLink",
    "DERControlResponse",
    "DERCurve",
    "DERCurveLink",
    "DERCurveList",
    "DERCurveListLink",
    "DERLink",
    "DERList",
    "DERListLink",
    "DERProgram",
    "DERProgramLink",
    "DERProgramList",
    "DERProgramListLink",
    "DERSettings",
    "DERSettingsLink",
    "DERStatus",
    "DERStatusLink",
    "DRLCCapabilities",
    "DateTimeInterval",
    "DefaultDERControl",
    "DefaultDERControlLink",
    "DemandResponseProgram",
    "DemandResponseProgramLink",
    "DemandResponseProgramList",
    "DemandResponseProgramListLink",
    "DeviceCapability",
    "DeviceCapabilityLink",
    "DeviceCategoryType",
    "DeviceInformation",
    "DeviceInformationLink",
    "DeviceStatus",
    "DeviceStatusLink",
    "DrResponse",
    "DutyCycle",
    "EndDevice",
    "EndDeviceControl",
    "EndDeviceControlList",
    "EndDeviceControlListLink",
    "EndDeviceLink",
    "EndDeviceList",
    "EndDeviceListLink",
    "EnvironmentalCost",
    "Error",
    "Event",
    "EventStatus",
    "File",
    "FileLink",
    "FileList",
    "FileListLink",
    "FileStatus",
    "FileStatusLink",
    "FixedPointType",
    "FixedVar",
    "FlowReservationRequest",
    "FlowReservationRequestList",
    "FlowReservationRequestListLink",
    "FlowReservationResponse",
    "FlowReservationResponseList",
    "FlowReservationResponseListLink",
    "FlowReservationResponseResponse",
    "FreqDroopType",
    "FunctionSetAssignments",
    "FunctionSetAssignmentsBase",
    "FunctionSetAssignmentsList",
    "FunctionSetAssignmentsListLink",
    "GPSLocationType",
    "HistoricalReading",
    "HistoricalReadingList",
    "HistoricalReadingListLink",
    "IEEE_802_15_4",
    "IPAddr",
    "IPAddrList",
    "IPAddrListLink",
    "IPInterface",
    "IPInterfaceList",
    "IPInterfaceListLink",
    "IdentifiedObject",
    "InverterStatusType",
    "LLInterface",
    "LLInterfaceList",
    "LLInterfaceListLink",
    "Link",
    "List_type",
    "ListLink",
    "LoadShedAvailability",
    "LoadShedAvailabilityList",
    "LoadShedAvailabilityListLink",
    "LocalControlModeStatusType",
    "LogEvent",
    "LogEventList",
    "LogEventListLink",
    "ManufacturerStatusType",
    "MessagingProgram",
    "MessagingProgramList",
    "MessagingProgramListLink",
    "MeterReading",
    "MeterReadingBase",
    "MeterReadingLink",
    "MeterReadingList",
    "MeterReadingListLink",
    "MirrorMeterReading",
    "MirrorMeterReadingList",
    "MirrorReadingSet",
    "MirrorUsagePoint",
    "MirrorUsagePointList",
    "MirrorUsagePointListLink",
    "Neighbor",
    "NeighborList",
    "NeighborListLink",
    "Notification",
    "NotificationList",
    "NotificationListLink",
    "Offset",
    "OperationalModeStatusType",
    "PEVInfo",
    "PowerConfiguration",
    "PowerFactor",
    "PowerFactorWithExcitation",
    "PowerStatus",
    "PowerStatusLink",
    "PrepayOperationStatus",
    "PrepayOperationStatusLink",
    "Prepayment",
    "PrepaymentLink",
    "PrepaymentList",
    "PrepaymentListLink",
    "PriceResponse",
    "PriceResponseCfg",
    "PriceResponseCfgList",
    "PriceResponseCfgListLink",
    "ProjectionReading",
    "ProjectionReadingList",
    "ProjectionReadingListLink",
    "RPLInstance",
    "RPLInstanceList",
    "RPLInstanceListLink",
    "RPLSourceRoutes",
    "RPLSourceRoutesList",
    "RPLSourceRoutesListLink",
    "RandomizableEvent",
    "RateComponent",
    "RateComponentLink",
    "RateComponentList",
    "RateComponentListLink",
    "ReactivePower",
    "ReactiveSusceptance",
    "Reading",
    "ReadingBase",
    "ReadingLink",
    "ReadingList",
    "ReadingListLink",
    "ReadingSet",
    "ReadingSetBase",
    "ReadingSetList",
    "ReadingSetListLink",
    "ReadingType",
    "ReadingTypeLink",
    "RealEnergy",
    "Registration",
    "RegistrationLink",
    "RequestStatus",
    "Resource",
    "RespondableIdentifiedObject",
    "RespondableResource",
    "RespondableSubscribableIdentifiedObject",
    "Response",
    "ResponseList",
    "ResponseListLink",
    "ResponseSet",
    "ResponseSetList",
    "ResponseSetListLink",
    "SelfDevice",
    "SelfDeviceLink",
    "ServiceChange",
    "ServiceSupplier",
    "ServiceSupplierLink",
    "ServiceSupplierList",
    "SetPoint",
    "SignedRealEnergy",
    "StateOfChargeStatusType",
    "StorageModeStatusType",
    "SubscribableIdentifiedObject",
    "SubscribableList",
    "SubscribableResource",
    "Subscription",
    "SubscriptionBase",
    "SubscriptionList",
    "SubscriptionListLink",
    "SupplyInterruptionOverride",
    "SupplyInterruptionOverrideList",
    "SupplyInterruptionOverrideListLink",
    "SupportedLocale",
    "SupportedLocaleList",
    "SupportedLocaleListLink",
    "TargetReading",
    "TargetReadingList",
    "TargetReadingListLink",
    "TargetReduction",
    "TariffProfile",
    "TariffProfileLink",
    "TariffProfileList",
    "TariffProfileListLink",
    "Temperature",
    "TextMessage",
    "TextMessageList",
    "TextMessageListLink",
    "TextResponse",
    "Time",
    "TimeConfiguration",
    "TimeLink",
    "TimeTariffInterval",
    "TimeTariffIntervalList",
    "TimeTariffIntervalListLink",
    "UnitValueType",
    "UnsignedFixedPointType",
    "UsagePoint",
    "UsagePointBase",
    "UsagePointLink",
    "UsagePointList",
    "UsagePointListLink",
    "VoltageRMS",
    "WattHour",
    "loWPAN",
]



from __future__ import annotations

import inspect
import logging
import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Literal, Optional, Tuple, Union

import yaml
from dataclasses_json import dataclass_json

__all__ = ["ServerConfiguration"]

#from ieee_2030_5.adapters import DERControlAdapter

try:
    from gridappsd.field_interface import MessageBusDefinition
except ImportError as ex:
    pass

import ieee_2030_5.models as m
from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.server.exceptions import NotFoundError
from ieee_2030_5.types_ import Lfdi

_log = logging.getLogger(__name__)

class InvalidConfigFile(Exception):
    pass


@dataclass
class DeviceConfiguration(m.EndDevice):
    # This id will be used to create certificates.
    id: str = None

    # This is used in the Registration function set but is
    # configured on the server and used to verify the correct location
    # on the client
    pin: int = None
    # hostname: str = None
    # ip: str = None
    # poll_rate: int = 900
    # # TODO: Direct control means that only one FSA will be available to the client.
    # direct_control: bool = True
    programs: List[str] = field(default_factory=list)
    
    ders: List[Dict] = field(default_factory=list)

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    def __hash__(self):
        return self.id.__hash__()


@dataclass
class DERCurveConfiguration(m.DERCurve):

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    def __hash__(self):
        return self.description.__hash__()


@dataclass
class DERControlBaseConfiguration(m.DERControlBase):

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    def __hash__(self):
        return self.description.__hash__()


@dataclass
class DERControlConfiguration(m.DERControl):
    description: str = None
    base: Dict = field(default_factory=dict)

    @classmethod
    def from_dict(cls, env):
        return cls(
            **{
                k: v
                for k, v in env.items() if k in inspect.signature(m.DERControl).parameters
                or k in inspect.signature(cls).parameters
            })

    def __hash__(self):
        return self.description.__hash__()


@dataclass
class DERProgramConfiguration(m.DERProgram):

    default_control: str = None
    controls: List[str] = field(default_factory=list)
    curves: List[str] = field(default_factory=list)

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    def __hash__(self):
        return self.description.__hash__()


@dataclass_json
@dataclass
class GridappsdConfiguration:
    field_bus_config: Optional[str] = None
    field_bus_def: Optional[MessageBusDefinition] = None
    feeder_id_file: Optional[str] = None
    feeder_id: Optional[str] = None
    simulation_id_file: Optional[str] = None
    simulation_id: Optional[str] = None


# @dataclass
# class ControlDefaults:
#     setESDelay: Optional[Int]
#     setESHighFreq: UInt16 [0..1]
#     setESHighVolt: Int16 [0..1]
#     setESLowFreq: UInt16 [0..1]
#     setESLowVolt: Int16 [0..1]
#     setESRampTms: UInt32 [0..1]
#     setESRandomDelay: UInt32 [0..1]
#     setGradW: UInt16 [0..1]
#     setSoftGradW: UInt16 [0..1]

# @dataclass
# class Program:
#     description: str
#     mRID: Optional[str]

# @dataclass
# class Control:
#     description: str
#     mRID: Optional[str]


@dataclass
class ProgramList:
    name: str
    programs: List[m.DERProgram]


@dataclass
class ServerConfiguration:
    openssl_cnf: str

    devices: List[DeviceConfiguration]

    tls_repository: str
    openssl_cnf: str

    server: str
    https_port: int

    log_event_list_poll_rate: int = 900
    device_capability_poll_rate: int = 900
    end_device_list_poll_rate: int = 86400  # daily check-in

    generate_admin_cert: bool = False

    http_port: int = None

    server_mode: Union[
        Literal["enddevices_create_on_start"],
        Literal["enddevices_register_access_only"]] = "enddevices_register_access_only"

    lfdi_mode: Union[
        Literal["lfdi_mode_from_file"],
        Literal["lfdi_mode_from_cert_fingerprint"]] = "lfdi_mode_from_cert_fingerprint"

    # Can include ip address as well
    server_hostname: str = None

    programs: List[DERProgramConfiguration] = field(default_factory=list)
    controls: List[DERControlConfiguration] = field(default_factory=list)
    curves: List[DERCurveConfiguration] = field(default_factory=list)
    events: List[Dict] = field(default_factory=list)

    # # map into program_lists array for programs for specific
    # # named list.
    # programs_map: Dict[str, int] = field(default_factory=dict)
    # program_lists: List[ProgramList] = field(default_factory=list)
    # fsa_list: List[FunctionSetAssignments] = field(default_factory=list)
    # curve_list: List[DERCurve] = field(default_factory=list)

    proxy_hostname: Optional[str] = None
    gridappsd: Optional[GridappsdConfiguration] = None
    # DefaultDERControl: Optional[DefaultDERControl] = None
    # DERControlList: Optional[DERControl] = field(default=list)

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    def __post_init__(self):
        self.server_hostname = f"{self.server}:{self.https_port}"
        self.curves = [DERCurveConfiguration.from_dict(x) for x in self.curves]
        self.controls = [DERControlConfiguration.from_dict(x) for x in self.controls]
        self.programs = [DERProgramConfiguration.from_dict(x) for x in self.programs]
        self.devices = [DeviceConfiguration.from_dict(x) for x in self.devices]
        for d in self.devices:
            d.deviceCategory = eval(f"m.DeviceCategoryType.{d.deviceCategory}").name
            #d.device_category_type = eval(f"m.DeviceCategoryType.{d.device_category_type}")

        # der_controls, der_default_control = None, None
        # if self.DERControlListFile:
        #     der_controls, der_default_control = DERControlAdapter.load_from_yaml_file(self.DERControlListFile)

        # temp_program_list = self.program_lists
        # if isinstance(self.program_lists, str):
        #     temp_program_list = yaml.safe_load(Path(self.program_lists).read_text())

        # self.program_lists = []
        # for program_list in temp_program_list['program_lists']:
        #     pl_obj = ProgramList(name=program_list["name"], programs=[])
        #     for inter_index, program_obj in enumerate(program_list['programs']):
        #         base = None
        #         if "DERControlBase" in program_obj:
        #             base = DERControlBase()
        #             for k in program_obj.get("DERControlBase"):
        #                 setattr(base, k, program_obj["DERControlBase"].get(k))
        #             del program_obj["DERControlBase"]
        #         # TODO Do Something with base so we can retrieve it.
        #         # if base:
        #         #     self.DefaultDERControl = DefaultDERControl(DERControlBase=base)
        #         #     for k, v in program.items():
        #         #         setattr(self.DefaultDERControl, k, v)
        #         # else:
        #         program = DERProgram()
        #
        #         for k, v in program_obj.items():
        #             setattr(program, k, v)
        #
        #         pl_obj.programs.append(program)
        #     self.program_lists.append(pl_obj)
        #
        # temp_curve_list = self.curve_list
        # if isinstance(self.curve_list, str):
        #     temp_curve_list = yaml.safe_load(Path(self.curve_list).read_text())
        #
        # self.curve_list = []
        # for item in temp_curve_list['curve_list']:
        #     curve_data: List[CurveData] = []
        #     for data in item.get('CurveData'):
        #         curve_data.append(CurveData(xvalue=data['xvalue'], yvalue=data['yvalue']))
        #     del item["CurveData"]
        #
        #     curve = DERCurve(CurveData=curve_data)
        #     for k, v in item.items():
        #         setattr(curve, k, v)
        #     self.curve_list.append(curve)

        if self.gridappsd:
            self.gridappsd = GridappsdConfiguration.from_dict(self.gridappsd)
            if Path(self.gridappsd.feeder_id_file).exists():
                self.gridappsd.feeder_id = Path(self.gridappsd.feeder_id_file).read_text().strip()
            if Path(self.gridappsd.simulation_id_file).exists():
                self.gridappsd.simulation_id = Path(
                    self.gridappsd.simulation_id_file).read_text().strip()

            if not self.gridappsd.feeder_id:
                raise ValueError(
                    "Feeder id from gridappsd not found in feeder_id_file nor was specified "
                    "in gridappsd config section.")

            # TODO: This might not be the best place for this manipulation
            self.gridappsd.field_bus_def = MessageBusDefinition.load(
                self.gridappsd.field_bus_config)
            self.gridappsd.field_bus_def.id = self.gridappsd.feeder_id

            _log.info("Gridappsd Configuration For Simulation")
            _log.info(f"feeder id: {self.gridappsd.feeder_id}")
            if self.gridappsd.simulation_id:
                _log.info(f"simulation id: {self.gridappsd.simulation_id}")
            else:
                _log.info("no simulation id")
            _log.info("x" * 80)

        # if self.field_bus_config:
        #     self.field_bus_def = MessageBusDefinition.load(self.field_bus_config)

    def get_device_pin(self, lfdi: Lfdi, tls_repo: TLSRepository) -> int:
        for d in self.devices:
            test_lfdi = tls_repo.lfdi(d.id)
            if test_lfdi == int(lfdi):
                return d.pin
        raise NotFoundError(f"The device_id: {lfdi} was not found.")
